# -*- coding: utf-8 -*-

"""Illustration feature.

This module implements the functionality to generate and save the
figures corresponding to the data stored in a container.

"""

from io import BytesIO
from logging import getLogger
from os.path import splitext

from matplotlib import font_manager
from matplotlib.pyplot import close, rcParams

from numlpa import config, kits
from numlpa.kits import fonts
from numlpa._version import version
from numlpa.core.parallelism import get_rank, get_size
from numlpa.core.storage import Container, deserialize


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.set_defaults(
        func=main,
    )
    parser.add_argument(
        'input_path',
        type=str,
        help="path of the input container",
    )
    parser.add_argument(
        'output_path',
        type=str,
        help="path of the output container",
    )
    parser.add_argument(
        '-f',
        '--format',
        type=str,
        choices=('pdf', 'svg', 'png'),
        help="output data files format",
    )
    parser.add_argument(
        '--font',
        type=str,
        choices=fonts.names(),
        help="text font used in the figures",
    )
    parser.add_argument(
        '--math',
        type=str,
        help="math font used in the figures",
        metavar='str',
    )
    parser.add_argument(
        '--size',
        type=int,
        help="font size used in the figures (pt)",
        metavar='int',
    )
    parser.add_argument(
        '--dpi',
        type=int,
        help="dots per inch",
        metavar='int',
    )

    logger.debug("defining command-line subparsers")
    subparsers = parser.add_subparsers(
        dest='representation',
        help="name of module used for illustration",
    )
    for name in kits.names('representations'):
        subparser = subparsers.add_parser(
            name=name,
            help=f"illustrate with {name} representation",
        )
        module = kits.get('representations', name)
        module.setup(subparser)


def main(input_path, output_path, **kwargs):
    """Generate figures.

    Parameters
    ----------
    input_path : str
        Path of the input container.
    output_path : str
        Path of the output container.

    Keyword Arguments
    -----------------
    representation : str
        Name of module used for illustration.
    format : str
        Output data files format.
    font : str
        Text font used in the figures.
    math : str
        Math font used in the figures.
    size : int
        Font size used in the figures (pt).
    dpi : int
        Dots per inch.

    """
    containers = {}
    files = {}
    modules = {}

    logger.debug("retrieving parameters")
    kwargs.setdefault('representation', config.get(__name__, 'representation'))
    kwargs.setdefault('format', config.get(__name__, 'format'))
    kwargs.setdefault('font', config.get(__name__, 'font'))
    kwargs.setdefault('math', config.get(__name__, 'math'))
    kwargs.setdefault('size', config.getint(__name__, 'size'))
    kwargs.setdefault('dpi', config.getint(__name__, 'dpi'))

    logger.debug("loading fonts")
    prop = font_manager.FontProperties(fname=fonts.path(kwargs['font']))
    font_manager.fontManager.addfont(fonts.path(kwargs['font']))
    rcParams["font.family"] = prop.get_name()
    rcParams["mathtext.fontset"] = kwargs['math']
    rcParams["font.size"] = kwargs['size']
    rcParams["legend.fontsize"] = kwargs['size']
    rcParams["axes.titlesize"] = kwargs['size']
    rcParams["figure.titlesize"] = kwargs['size']

    logger.debug("preparing containers")
    containers['input'] = Container(input_path)
    containers['output'] = Container(output_path, create=True)

    logger.debug("preparing file names")
    files['input'] = containers['input'].names()
    files['stems'] = [splitext(name)[0] for name in files['input']]
    files['output'] = [f"{stem}.{kwargs['format']}" for stem in files['stems']]
    files['existing'] = containers['output'].names()

    logger.debug("preparing submodules")
    modules['repr'] = kits.get('representations', kwargs['representation'])

    logger.debug("generating files")
    static = {
        'containers': containers,
        'files': files,
        'kwargs': kwargs,
        'modules': modules,
    }
    for i in range(get_rank(), len(files['input']), get_size()):
        process(static, i)


def process(static, i):
    """Process a file.

    Parameters
    ----------
    static: dict of dict
        The data that does not vary from one file to another.
    i : int
        Index of the file to be generated.

    Raises
    ------
    TypeError
        If the input data format is not recognized.

    """
    data = {}
    file = {}
    passed = {}
    series = {}

    logger.debug("defining local variables for file index %d", i)
    file['input'] = static['files']['input'][i]
    file['output'] = static['files']['output'][i]
    file['target'] = f"({static['containers']['output']})[{file['output']}]"
    passed.update(static['kwargs'])

    logger.debug("checking if %s already exists", file['output'])
    if file['output'] in static['files']['existing']:
        logger.debug("skipping %s", file['output'])
        return

    logger.debug("loading %s", file['input'])
    series['input'] = static['containers']['input'].get(file['input'])
    data['input'] = deserialize(series['input'], file['input'])

    logger.debug("checking %s", file['input'])
    try:
        data['type'] = data['input']['metadata']['type']
    except (KeyError, TypeError) as exc:
        raise RuntimeError(f"can not parse data from {file['input']}") from exc
    if data['type'] == 'dislocation-sample':
        data['function'] = 'sample'
        data['title'] = "Dislocation map"
    elif data['type'] == 'fourier-transform':
        data['function'] = 'transform'
        data['title'] = "Diffraction profile"
    elif data['type'] == 'model-adjustment':
        data['function'] = 'adjustment'
        data['title'] = "Model adjustment"
    else:
        raise TypeError(f"can not represent data type {data['type']}")
    if hasattr(static['modules']['repr'], data['function']):
        generator = getattr(static['modules']['repr'], data['function'])
    else:
        logger.warning(
            "module '%s' can not be used to represent '%s' data, "
            "switched to 'basic' representation module",
            static['modules']['repr'].__name__,
            data['type'],
        )
        static['modules']['repr'] = kits.get('representations', 'basic')
        generator = getattr(static['modules']['repr'], data['function'])

    logger.info("generating %s", file['target'])
    data['output'] = {}
    data['output']['fig'] = generator(data['input'], **passed)
    data['output']['metadata'] = {
        'Title': data['title'],
        'Creator': f'NumLPA {version}',
    }
    series['buffer'] = BytesIO()
    data['output']['fig'].savefig(
        series['buffer'],
        format=passed['format'],
        dpi=passed['dpi'],
        metadata=data['output']['metadata'],
    )
    close(data['output']['fig'])
    series['buffer'].seek(0)

    logger.debug("saving %s", file['output'])
    series['output'] = bytes(series['buffer'].read())
    static['containers']['output'].add(series['output'], file['output'])



# -*- coding: utf-8 -*-

"""CPU diffractometer.

This implementation uses the Monte Carlo method to simulate X-ray
diffraction and is parallelized on the different cores of the CPU.

"""

from . import diffraction, parser


setup = parser.setup
diffract = diffraction.diffract



# -*- coding: utf-8 -*-

"""CPU spatial analyzer.

"""

from logging import getLogger


logger = getLogger(__package__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.add_argument(
        '--type',
        help="dislocation type",
        type=str,
        choices=('screw', 'edge'),
    )
    parser.add_argument(
        '--r_0',
        help="core radius (m)",
        type=float,
        metavar='float',
    )
    parser.add_argument(
        '--r_roi',
        help="size of the region of interest (m)",
        type=float,
        metavar='float',
    )
    parser.add_argument(
        '--r_max',
        help="maximum radius value (m)",
        type=float,
        metavar='float',
    )
    parser.add_argument(
        '--steps',
        help="number of subdivision on the absice axis",
        type=int,
        metavar='int',
    )
    parser.add_argument(
        '--b_len',
        help="Burgers vector length (m)",
        type=float,
        metavar='float',
    )
    parser.add_argument(
        '--poisson',
        help="Poisson number",
        type=float,
        metavar='float',
    )
    parser.add_argument(
        '--shear',
        help="shear modulus (Pa)",
        type=float,
        metavar='float',
    )
    parser.add_argument(
        '--processes',
        help="number of parallel processes",
        type=int,
        metavar='int',
    )



# -*- coding: utf-8 -*-

"""Affine restriction.

This module restricts the interval of the Fourier variable to the part
where the transformed amplitude :math:`\\log[A(L)]/L^2` follows an
affine behavior when viewed with the Fourier variable axis in
logarithmic scale. An example of the application of this restriction is
given below:

.. image:: ../figures/restrictions/affine.svg
   :width: 600

"""

import numpy as np


def limit(variable, amplitude):
    """Return the number of points to keep.

    Parameters
    ----------
    variable : list of float
        Fourier variable values.
    amplitudes : list of float
        Amplitude values.

    Returns
    -------
    int
        Number of points to keep.

    """
    n_positive = 1
    while n_positive < len(variable) and amplitude[n_positive] > 0:
        n_positive += 1
    variable = variable[:n_positive]
    amplitude = amplitude[:n_positive]
    x_list = np.log(variable)
    y_list = np.log(amplitude) / np.array(variable)**2
    n_p = n_affine(x_list, y_list)
    return n_p


def n_affine(x_list, y_list, tol=0.005):
    """Return the number of points to keep.

    Parameters
    ----------
    x_list : list of float
        Fourier variable values after transformation.
    y_list : list of float
        Amplitude values after transformation.

    Returns
    -------
    int
        Number of points to keep.

    """
    n_p = 3
    stop = False
    while n_p < len(x_list) and not stop:
        n_p += 1
        weights = np.ones(n_p)
        weighted = np.stack((x_list[:n_p], weights)).T
        error = np.linalg.lstsq(weighted, y_list[:n_p], rcond=None)[1][0]
        deviation = np.sqrt(error/(n_p-1))
        if deviation/np.ptp(y_list[:n_p]) > tol:
            stop = True
    return n_p - 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Restriction figure generator.

"""

from matplotlib import font_manager
from matplotlib.pyplot import rcParams
import matplotlib.pyplot as plt
import numpy as np

from numlpa import config, kits
from numlpa.kits import fonts


font = config.get('numlpa.main.export', 'font')
math = config.get('numlpa.main.export', 'math')
prop = font_manager.FontProperties(fname=fonts.path(font))
font_manager.fontManager.addfont(fonts.path(font))
rcParams["font.family"] = prop.get_name()
rcParams["mathtext.fontset"] = math


def export_basic(restriction, file, variable, amplitude):
    """Export a figure to illustrate the restriction.

    """
    fig, ax = plt.subplots()
    ax.plot(variable, amplitude, "o-", label="simulated $A(L)$")
    i = restriction.limit(variable, amplitude)
    ax.plot(variable[:i], amplitude[:i], ".-", label="restricted $A(L)$")
    ax.grid()
    ax.legend()
    ax.set_xlabel("$L$")
    fig.savefig(file)


def export_affine(restriction, file, variable, amplitude):
    """Export a figure to illustrate the restriction.

    """
    i = restriction.limit(variable, amplitude)
    fig, ax = plt.subplots()
    ax.set_xscale('log')
    amplitude = np.log(amplitude)/variable**2
    ax.plot(variable, amplitude, "o-", label="simulated $\ln[A(L)]/L^2$")
    ax.plot(variable[:i], amplitude[:i], label="restricted $\ln[A(L)]/L^2$")
    ax.grid()
    ax.legend()
    ax.set_xlabel("$L$")
    fig.savefig(file)


if __name__ == '__main__':

    export_affine(
        kits.get('restrictions', 'affine'),
        'affine.svg',
        np.exp(np.array([0, 1, 2, 3, 4, 5, 6, 7])),
        np.exp(
            np.exp(np.array([0, 1, 2, 3, 4, 5, 6, 7]))**2 *
            (np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.2, 6.8, 8.6])/100000-0.0002),
        )
    )

    export_basic(
        kits.get('restrictions', 'none'),
        'none.svg',
        np.array([0, 1, 2, 3, 4, 5, 6, 7]),
        np.array([10, 9.5, 8.5, 6.5, 2.5, 3, -2, -1])/10,
    )

    export_basic(
        kits.get('restrictions', 'positive'),
        'positive.svg',
        np.array([0, 1, 2, 3, 4, 5, 6, 7]),
        np.array([10, 9.5, 8.5, 6.5, 2.5, 3, -2, -1])/10,
    )



# -*- coding: utf-8 -*-

"""Positive restriction.

This module allows to restrict the interval of the Fourier variable to
the part where the amplitude is strictly positive. An example of the
application of this restriction is given below:

.. image:: ../figures/restrictions/positive.svg
   :width: 600

"""


def limit(variable, amplitude):
    """Return the number of points to keep.

    Parameters
    ----------
    variable : list of float
        Fourier variable values.
    amplitudes : list of float
        Amplitude values.

    Returns
    -------
    int
        Number of points to keep.

    """
    n_positive = 0
    while n_positive < len(variable) and amplitude[n_positive] > 0:
        n_positive += 1
    return n_positive



# -*- coding: utf-8 -*-

"""Writing and reading of data.

This module implements the functionalities for writing and reading
files in the computer permanent storage.

"""

import tarfile
import zipfile
import json
import pickle
from io import BytesIO
from logging import getLogger
from os import getpid, listdir, mkdir, remove
from os.path import exists, isdir, join
from signal import signal, SIGINT, SIGTERM
from sys import exit as sys_exit
from time import time


logger = getLogger(__name__)


class Container:
    """File container handler.

    This class implements a single interface for writing and reading
    files in a tarball, a zip archive, or a directory.

    Attributes
    ----------
    path : str
        Path to the container.
    format : str
        Container type ('tar', 'zip' or 'dir').
    killed : bool
        Set to True when a stop signal is received during a write.
    signum : int
        Stop signal number.
    lockfile : str
        Path of the locking file.

    """

    def __init__(self, path, **kwargs):
        """Initialize the container handler.

        Parameters
        ----------
        path : str
            Path to the container.

        Keyword Arguments
        -----------------
        create : bool
            Set to True if the container can be created.

        """
        logger.debug("instantiating container handler %s", path)
        self.path = path
        if path.endswith('.tar'):
            self.format = 'tar'
        elif path.endswith('.zip'):
            self.format = 'zip'
        else:
            self.format = 'dir'
        self.killed = False
        self.signum = None
        self.lockfile = f'{self.path}.lock'
        if kwargs.get('create', False):
            self.create()
        self.check()

    def __str__(self):
        """Return a string representation of the container.

        Returns
        -------
        str
            String representation of the container.

        """
        return self.path

    def create(self):
        """Create the container in the file system.

        """
        logger.debug("creating container %s", self)
        try:
            if self.format == 'tar':
                with tarfile.open(self.path, 'x'):
                    pass
            elif self.format == 'zip':
                with zipfile.ZipFile(self.path, 'x'):
                    pass
            elif self.format == 'dir':
                mkdir(self.path)
            logger.info("created container %s", self)
        except FileExistsError:
            pass

    def check(self):
        """Ensure that the container exists and is valid.

        Raises
        ------
        FileNotFoundError
            If the container is not found at the given path.
        FileExistsError
            If the path does not correspond to a valid container.

        """
        logger.debug("checking container %s", self)
        if not exists(self.path):
            raise FileNotFoundError(f"container '{self}' not found")
        if self.format == 'tar':
            if not tarfile.is_tarfile(self.path):
                raise FileExistsError(f"'{self}' is not a valid tar file")
        elif self.format == 'zip':
            if not zipfile.is_zipfile(self.path):
                raise FileExistsError(f"'{self}' is not a valid zip file")
        elif self.format == 'dir':
            if not isdir(self.path):
                raise FileExistsError(f"'{self}' is not a directory")

    def names(self):
        """Return the name of the files stored in the container.

        Returns
        -------
        list of str
            Name of the files stored in the container.

        """
        logger.debug("retrieving file names in %s", self)
        if self.format == 'tar':
            with tarfile.open(self.path, 'r') as archive:
                listing = archive.getnames()
        elif self.format == 'zip':
            with zipfile.ZipFile(self.path, 'r') as archive:
                listing = archive.namelist()
        elif self.format == 'dir':
            listing = listdir(self.path)
        return sorted(listing)

    def lock(self):
        """Lock the container.

        This function allows to prevent another container handler
        object from placing a file in the container. This feature is a
        solution to the concurrent write problem when the file
        processing is parallelized.

        See Also
        --------
        unlock

        """
        logger.debug("locking %s", self)
        locked = False
        while not locked:
            try:
                with open(self.lockfile, 'x', encoding='utf-8') as file:
                    file.write(str(getpid()))
                locked = True
            except IOError:
                pass

    def unlock(self):
        """Unlock the container.

        See Also
        --------
        lock

        """
        logger.debug("unlocking %s", self)
        remove(self.lockfile)

    def add(self, series, name):
        """Add a file in the container.

        Parameters
        ----------
        series : bytes
            File data.
        name : str
            File name.

        """
        logger.debug("adding %s to %s", name, self)
        sigint = signal(SIGINT, self.handler)
        sigterm = signal(SIGTERM, self.handler)
        if self.format == 'tar':
            self.lock()
            with tarfile.open(self.path, 'a') as archive:
                info = tarfile.TarInfo(name)
                info.mtime = time()
                info.size = len(series)
                archive.addfile(info, BytesIO(series))
            self.unlock()
        elif self.format == 'zip':
            self.lock()
            with zipfile.ZipFile(self.path, 'a') as archive:
                archive.writestr(name, series)
            self.unlock()
        elif self.format == 'dir':
            with open(join(self.path, name), 'wb') as file:
                file.write(series)
        if self.killed:
            if self.signum == SIGINT:
                raise KeyboardInterrupt
            sys_exit(143)
        signal(SIGINT, sigint)
        signal(SIGTERM, sigterm)

    def iterator(self, names):
        """Return an iterator loading the data of the requested files.

        Parameters
        ----------
        names : list of str
            Names of the files to retrieve.

        Yields
        ------
        bytes
            File data.

        """
        logger.debug("iterating over %s", self)
        if self.format == 'tar':
            with tarfile.open(self.path, 'r') as archive:
                for member in names:
                    buffer = archive.extractfile(member)
                    yield bytes(buffer.read())
        elif self.format == 'zip':
            with zipfile.ZipFile(self.path, 'r') as archive:
                for member in names:
                    yield archive.read(member)
        elif self.format == 'dir':
            for member in names:
                with open(join(self.path, member), mode='rb') as file:
                    yield file.read()

    def get(self, name):
        """Return the file data.

        Parameters
        ----------
        name : str
            Name of the file to retrieve.

        Returns
        -------
        bytes
            File data.

        """
        return next(self.iterator([name]))

    def handler(self, signum, frame=None):
        """Custom signal handler.

        Parameters
        ----------
        signum : int
            Signal number.
        frame : frame object
            Current stack frame.

        """
        logger.debug("handling signal %d", signum)
        self.killed = True
        _ = frame
        self.signum = signum


def serialize(data, name):
    """Return the serialized data.

    Parameters
    ----------
    data : dict
        Data to be serialized.
    name : str
        Serialization format or file name.

    Returns
    -------
    bytes
        Serialized data.

    Raises
    ------
    RuntimeError
        If the serialization format is not recognized.

    See Also
    --------
    deserialize

    """
    logger.debug("serializing %s", name)
    if name.endswith('json'):
        return json.dumps(data).encode('utf-8')
    if name.endswith('pyc'):
        return pickle.dumps(data)
    raise RuntimeError(f"invalid serialization format for '{name}'")


def deserialize(series, name):
    """Return the deserialized data.

    Parameters
    ----------
    series : bytes
        Data to be deserialized.
    name : str
        Serialization format or file name.

    Returns
    -------
    dict
        Deserialized data.

    Raises
    ------
    RuntimeError
        If the serialization format is not recognized.

    See Also
    --------
    serialize

    """
    logger.debug("deserializing %s", name)
    if name.endswith('json'):
        return json.loads(series.decode('utf-8'))
    if name.endswith('pyc'):
        return pickle.loads(series)
    raise RuntimeError(f"invalid deserialization format for '{name}'")



# -*- coding: utf-8 -*-

"""Dipoles dislocation probability distribution.

This distribution allows to generates pairs of dislocations of opposite
sign and separated by a fixed distance. The region of interest is
divided into subareas of equal size in which a fixed number of dipoles
are generated.

Below is an example of a set of parameter values for the distribution
and an illustration of a resulting dislocation sample.

* ``length`` = ``0`` (The dipole length then becomes the average
  interdislocation distance.)

* ``nsub`` = ``5`` (A side of the region of interest is divided into 5
  subareas.)

* ``pairs`` = ``4`` (Each subarea contains 4 dipoles of dislocations.)

* ``side`` = ``2000e-9`` (The region of interest is a 2000 nm square.)

.. image:: ../figures/distributions/dipoles/0.svg
   :width: 600

"""

from logging import getLogger
from math import cos, sin, pi, sqrt
from random import seed as set_seed, uniform

from numlpa import config


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.add_argument(
        '--length',
        type=float,
        help="distance between two dislocations of a dipole (m)",
        metavar='float',
    )
    parser.add_argument(
        '--nsub',
        type=int,
        help="number of subareas on a side of the region of interest",
        metavar='int',
    )
    parser.add_argument(
        '--pairs',
        type=int,
        help="number of dipoles in each subarea",
        metavar='int',
    )
    parser.add_argument(
        '--side',
        type=float,
        help="size of the region of interest (m)",
        metavar='float',
    )


def draw(seed, **kwargs):
    """Return a sample drawn from the dipoles distribution.

    When the 'length' parameter is zero, the dipole distance becomes
    the average interdislocation distance.

    Parameters
    ----------
    seed : int
        Random seed.

    Keyword Arguments
    -----------------
    length : float
        Distance between two dislocations of a dipole (m).
    nsub : int
        Number of subareas on a side of the region of interest.
    pairs : int
        Number of dipoles in each subarea.
    side : float
        Size of the region of interest (m).

    Returns
    -------
    dict
        Dislocation sample data.

    """
    logger.debug("retrieving parameters")
    kwargs.setdefault('length', config.getfloat(__name__, 'length'))
    kwargs.setdefault('nsub', config.getint(__name__, 'nsub'))
    kwargs.setdefault('pairs', config.getint(__name__, 'pairs'))
    kwargs.setdefault('side', config.getfloat(__name__, 'side'))
    density = 2*kwargs['pairs']*kwargs['nsub']**2/kwargs['side']**2
    subarea = kwargs['side']/kwargs['nsub']
    if kwargs['length'] == 0:
        kwargs['length'] = 1/sqrt(density)

    logger.debug("drawing dislocations")
    set_seed(seed)
    x_list, y_list, s_list = [], [], []
    for i in range(kwargs['nsub']):
        for j in range(kwargs['nsub']):
            for _ in range(kwargs['pairs']):
                x_center = uniform(i*subarea, (i+1)*subarea)
                y_center = uniform(j*subarea, (j+1)*subarea)
                angle = uniform(0, 2*pi)
                for sense in (-1, 1):
                    x_list.append(x_center+sense*cos(angle)*kwargs['length']/2)
                    y_list.append(y_center+sense*sin(angle)*kwargs['length']/2)
                    s_list.append(sense)

    logger.debug("assembling data")
    data = {
        'distribution': {
            'module': __name__,
            'seed': seed,
            'density': density,
            'length': kwargs['length'],
            'nsub': kwargs['nsub'],
            'pairs': kwargs['pairs'],
            'side': kwargs['side'],
        },
        'region': {
            'type': 'square',
            'side': kwargs['side'],
        },
        'dislocations': {
            'senses': s_list,
            'positions': [x_list, y_list],
        },
    }
    return data



# -*- coding: utf-8 -*-

"""No restriction.

This module does not apply any restriction on the Fourier variable
interval for model fitting. As shown below, the restricted interval is
equal to the original interval:

.. image:: ../figures/restrictions/none.svg
   :width: 600

"""


def limit(variable, amplitude):
    """Return the number of points to keep.

    Parameters
    ----------
    variable : list of float
        Fourier variable values.
    amplitudes : list of float
        Amplitude values.

    Returns
    -------
    int
        Number of points to keep.

    """
    _ = amplitude
    return len(variable)



# -*- coding: utf-8 -*-

"""Numerical Line Profile Analysis package.

This Python package is intended for the analysis of crystal
microstructure characterization models from simulated X-ray
diffraction line profiles.

"""

__author__ = 'Dunstan Becht'

from configparser import ConfigParser
from logging import basicConfig
from pkgutil import get_data


config = ConfigParser()
config.read_string(get_data(__package__, 'default.conf').decode())

basicConfig(level=config.get(__name__, 'log'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Test script.

This script executes the unit tests for the package.

"""

from contextlib import contextmanager
from json import dumps, load, loads
from pathlib import Path
from shutil import copytree
from subprocess import CalledProcessError, run
from sys import stderr
from tempfile import TemporaryDirectory, TemporaryFile
from unittest import main, TestCase

from numlpa import kits


dir_tests = Path(__file__).parent
dir_material = dir_tests/'material'

with open(dir_tests/'references.json', 'r', encoding='utf-8') as ref_file:
    ref = load(ref_file)


def runcheck(command):
    """Run a command and check the exit code.

    If the exit code of the executed command is different from 0, a
    RuntimeError containing the output of the command is raised.

    Parameters
    ----------
    command : str
        Command to be runned.

    Raises
    ------
    RuntimeError
        If the command exited with an error code.

    """
    with TemporaryFile() as file:
        try:
            run(command, shell=True, check=True, stdout=file, stderr=file)
        except CalledProcessError as error:
            file.seek(0)
            raise RuntimeError(file.read().decode()) from error


@contextmanager
def temporary(directory='.'):
    """Return a temporary clone of the target test directory.

    Parameters
    ----------
    directory : str
        Name of the directory to clone from material.

    Returns
    -------
    str
        Path to the cloned target directory.

    """
    temp_dir = TemporaryDirectory()
    path = Path(temp_dir.name)
    copytree(dir_material/directory, path, dirs_exist_ok=True)
    try:
        yield path
    finally:
        temp_dir.cleanup()


def display(data):
    """Print data in JSON format.

    Parameters
    ----------
    data : dict or list
        Data to be displayed.

    """
    print("\n"+dumps(rounded(data), indent=4), file=stderr)


def rounded(data):
    """Decreases the float precision.

    Transcendental functions can have very close but different results
    for different implementations. The goal here is to hide the part
    of the number that may be different for different implementations.

    Parameters
    ----------
    data
        JSON serializable data.

    Returns
    -------
    Any
        Data with decreased precision.

    """
    def parse_float(number):
        return float(f'{float(number):1.7e}')
    return loads(dumps(data, indent=4), parse_float=parse_float)


class TestKits(TestCase):
    """Class for testing kits."""

    def test_distributions(self):
        """Test dislocation distributions."""
        self.__dict__['maxDiff'] = None
        tested = kits.names('distributions')
        for name in tested:
            with self.subTest(name=name):
                distribution = kits.get('distributions', name)
                parameters = ref['distributions'][name]['parameters']
                dat = distribution.draw(**parameters)
                self.assertEqual(
                    ref['distributions'][name]['returns']['positions'],
                    rounded(dat['dislocations']['positions']),
                )
                self.assertEqual(
                    ref['distributions'][name]['returns']['senses'],
                    dat['dislocations']['senses'],
                )
                self.assertEqual(
                    ref['distributions'][name]['returns']['density'],
                    rounded(dat['distribution']['density']),
                )

    def test_diffractometers(self):
        """Test diffractometers."""
        self.__dict__['maxDiff'] = None
        tested = kits.names('diffractometers')
        for name in tested:
            for kind in ('screw', 'edge',):
                with self.subTest(name=name, kind=kind):
                    params_key = f'parameters-{kind}'
                    return_key = f'returns-{kind}'
                    diffractometer = kits.get('diffractometers', name)
                    parameters = ref['diffractometers'][name][params_key]
                    dat = diffractometer.diffract(**parameters)
                    self.assertEqual(
                        ref['diffractometers'][name][return_key],
                        rounded(dat['coefficients']),
                    )

    def test_models(self):
        """Test LPA models."""
        self.__dict__['maxDiff'] = None
        tested = kits.names('models')
        for name in tested:
            transform = ref['models'][name]['parameters']['transform']
            harmonic = ref['models'][name]['parameters']['harmonic']
            parameters = ref['models'][name]['parameters']['parameters']
            model = kits.get('models', name)
            dat = model.model(transform, harmonic)(*parameters)
            self.assertEqual(
                ref['models'][name]['returns'],
                rounded(list(dat)),
            )


class TestCommandLineInterface(TestCase):
    """Class for testing the command-line interface."""

    def test_draw(self):
        """Test draw command."""
        for distribution in kits.names('distributions'):
            with self.subTest(distribution=distribution):
                with temporary() as target:
                    runcheck(f"cd {target}; numlpa draw {distribution}")

    def test_diffract(self):
        """Test diffract command."""
        for diff in kits.names('diffractometers'):
            with self.subTest(diff=diff):
                with temporary() as target:
                    runcheck(f"cd {target}; numlpa draw samples -n 1")
                    runcheck(f"cd {target}; numlpa diffract samples {diff}")


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Data extraction feature.

This module implements the functionality that allows you to easily and
quickly extract specific data from files saved in a container.

"""

from json import dumps
from logging import getLogger

import numpy as np

from numlpa import config
from numlpa.core.parallelism import get_rank
from numlpa.core.storage import Container, deserialize, serialize


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.set_defaults(
        func=main,
    )
    parser.add_argument(
        'input_path',
        type=str,
        help="path of the input container",

    )
    parser.add_argument(
        '-e',
        '--entry',
        type=str,
        help="path to the field to be extracted (e.g. 'metadata.version')",
        metavar='key1.key2...',
    )
    parser.add_argument(
        '-o',
        '--output_path',
        type=str,
        help="path to the output file",
        metavar='path',
    )
    parser.add_argument(
        '-f',
        '--format',
        type=str,
        choices=('json', 'pyc'),
        help="format of the output file",
    )
    selection = parser.add_mutually_exclusive_group()
    selection.add_argument(
        '--first',
        type=int,
        help="select value from the first 'n' files (in alphabetical order)",
        metavar='int',
    )
    selection.add_argument(
        '--last',
        type=int,
        help="select value from the last 'n' files (in alphabetical order)",
        metavar='int',
    )
    treatment = parser.add_mutually_exclusive_group()
    treatment.add_argument(
        '--mean',
        action='store_true',
        help="compute the average of the selected values",
    )
    treatment.add_argument(
        '--deviation',
        action='store_true',
        help="compute the standard deviation of the selected values",
    )


def main(input_path, **kwargs):
    """Extract data from the container.

    Parameters
    ----------
    input_path : str
        Path of the input container.

    Keyword Arguments
    -----------------
    entry : str
        Path to the field to be extracted (e.g. 'metadata.version').
    output : str
        Path to the output file.
    format : str
        Format of the output file.
    first : int
        Select value from the first 'n' files (in alphabetical order).
    last : int
        Select value from the last 'n' files (in alphabetical order).
    mean : bool
        Compute the average of the selected values.
    deviation : bool
        Compute the standard deviation of the selected values.

    """
    containers = {}
    data = {}
    if get_rank() != 0:
        return

    logger.debug("retrieving parameters")
    kwargs.setdefault('format', config.get(__name__, 'format'))
    kwargs.setdefault('first', 0)
    kwargs.setdefault('last', 0)
    kwargs.setdefault('mean', False)
    kwargs.setdefault('deviation', False)

    logger.debug("preparing containers")
    containers['input'] = Container(input_path)
    if 'output_path' in kwargs:
        containers['output'] = Container(kwargs['output_path'], create=True)
    else:
        containers['output'] = None

    logger.debug("checking entry parameter")
    if 'entry' not in kwargs:
        retrieve_entries(containers, data)
        print("available entries:")
        for data_entry, data_type in data['entries']:
            print(f"- {data_entry} ({data_type.__name__})")
        return

    logger.debug("retrieving data")
    retrieve_values(containers, data, kwargs)
    if not containers['output']:
        print(dumps(data['output'], indent=4))


def retrieve_values(containers, data, kwargs):
    """Retrieve data.

    Parameters
    ----------
    containers : dict of Container
        Input and output containers.
    data : dict
        Dictionary in which to place the extracted data.
    kwargs : dict
        Keyword arguments passed to the main function.

    """
    files = {}
    series = {}

    logger.debug("extracting file data")
    data['output'] = []
    files['input'] = containers['input'].names()
    if kwargs['first']:
        files['input'] = files['input'][:kwargs['first']]
    elif kwargs['last']:
        files['input'] = files['input'][-kwargs['last']:]
    series['iterator'] = containers['input'].iterator(files['input'])
    for i in range(len(files['input'])):
        series['input'] = next(series['iterator'])
        data['input'] = deserialize(series['input'], files['input'][i])
        for node in kwargs['entry'].split('.'):
            if isinstance(data['input'], list):
                node = int(node)
            data['input'] = data['input'][node]
        data['output'].append(data.pop('input'))

    logger.debug("processing data")
    if kwargs['mean']:
        data['output'] = np.array(data['output'], dtype=np.float64)
        data['output'] = np.mean(data['output'], axis=0)
    elif kwargs['deviation']:
        data['output'] = np.array(data['output'], dtype=np.float64)
        data['output'] = np.std(data['output'], axis=0)
    if isinstance(data['output'], np.ndarray):
        data['output'] = np.where(
            np.isnan(data['output']),
            None,
            data['output'],
        ).tolist()

    logger.debug("saving data")
    if not containers['output']:
        return
    files['output'] = f"extracted.{kwargs['format']}"
    files['target'] = f"({containers['output']})[{files['output']}]"
    files['existing'] = containers['output'].names()
    if files['output'] in files['existing']:
        logger.debug("skipping %s", files['output'])
        return
    logger.info("generating %s", files['target'])
    series['output'] = serialize(data['output'], files['output'])
    containers['output'].add(series['output'], files['output'])


def retrieve_entries(containers, data):
    """Get available entries from the container.

    Parameters
    ----------
    containers : dict of Container
        Input and output containers.
    data : dict
        Dictionary in which to place the extracted data.

    """
    series = {}
    files = {}

    logger.debug("extracting first file")
    files['input'] = containers['input'].names()
    series['input'] = containers['input'].get(files['input'][0])
    data['input'] = deserialize(series['input'], files['input'][0])

    def rec(node, parents):
        if isinstance(node, dict):
            entries = []
            for child in node:
                entries += rec(node[child], parents+[child])
            return entries
        return [('.'.join(parents), type(node))]

    logger.debug("retrieving values")
    data['entries'] = rec(data['input'], [])



# -*- coding: utf-8 -*-

"""Wilkens dislocation probability distribution.

The Wilkens distribution divides the region of interest into subareas
of equal size and randomly places in these subareas as many
dislocations with a positive Burger vector sense as negative.

Below is an example of a set of parameter values for the distribution
and an illustration of a resulting dislocation sample.

* ``nsub`` = ``5`` (A side of the region of interest is divided into 5
  subareas.)

* ``pairs`` = ``4`` (Each subarea contains 8 dislocations.)

* ``side`` = ``2000e-9`` (The region of interest is a 2000 nm square.)

.. image:: ../figures/distributions/wilkens/0.svg
   :width: 600

"""

from logging import getLogger
from random import seed as set_seed, uniform

from numlpa import config


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.add_argument(
        '--nsub',
        type=int,
        help="number of subareas on a side of the region of interest",
        metavar='int',
    )
    parser.add_argument(
        '--pairs',
        type=int,
        help="number of pairs of dislocations in each subarea",
        metavar='int',
    )
    parser.add_argument(
        '--side',
        type=float,
        help="size of the region of interest (m)",
        metavar='float',
    )


def draw(seed, **kwargs):
    """Return a sample drawn from the Wilkens distribution.

    Parameters
    ----------
    seed : int
        Random seed.

    Keyword Arguments
    -----------------
    nsub : int
        Number of subareas on a side of the region of interest.
    pairs : int
        Number of pairs of dislocations in each subarea.
    side : float
        Size of the region of interest (m).

    Returns
    -------
    dict
        Dislocation sample data.

    """
    logger.debug("retrieving parameters")
    kwargs.setdefault('nsub', config.getint(__name__, 'nsub'))
    kwargs.setdefault('pairs', config.getint(__name__, 'pairs'))
    kwargs.setdefault('side', config.getfloat(__name__, 'side'))
    density = 2*kwargs['pairs']*kwargs['nsub']**2/kwargs['side']**2
    subarea = kwargs['side']/kwargs['nsub']

    logger.debug("drawing dislocations")
    set_seed(seed)
    x_list, y_list, s_list = [], [], []
    for i in range(kwargs['nsub']):
        for j in range(kwargs['nsub']):
            for _ in range(kwargs['pairs']):
                for sense in (-1, 1):
                    x_list.append(uniform(i*subarea, (i+1)*subarea))
                    y_list.append(uniform(j*subarea, (j+1)*subarea))
                    s_list.append(sense)

    logger.debug("assembling data")
    data = {
        'distribution': {
            'module': __name__,
            'seed': seed,
            'density': density,
            'nsub': kwargs['nsub'],
            'pairs': kwargs['pairs'],
            'side': kwargs['side'],
        },
        'region': {
            'type': 'square',
            'side': kwargs['side'],
        },
        'dislocations': {
            'senses': s_list,
            'positions': [x_list, y_list],
        },
    }
    return data



# -*- coding: utf-8 -*-

"""Data fusion feature.

This module implements the functionality to merge data from several
similar files into a single file.

"""

from datetime import datetime, timezone
from logging import getLogger
from os.path import splitext

import numpy as np

from numlpa import config
from numlpa._version import version_tuple
from numlpa.core.parallelism import get_rank
from numlpa.core.storage import Container, deserialize, serialize


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.set_defaults(
        func=main,
    )
    parser.add_argument(
        'input_path',
        type=str,
        help="path of the input container",
    )
    parser.add_argument(
        'output_path',
        type=str,
        help="path of the output container",
    )
    parser.add_argument(
        '-f',
        '--format',
        type=str,
        choices=('json', 'pyc'),
        help="format of the output files",
    )


def main(input_path, output_path, **kwargs):
    """Merge the files.

    Parameters
    ----------
    input_path : str
        Path of the input container.
    output_path : str
        Path of the output container.

    Keyword Arguments
    -----------------
    format : str
        Format of the output file.

    """
    containers = {}
    data = {}
    files = {}
    series = {}
    if get_rank() != 0:
        return

    logger.debug("retrieving parameters")
    kwargs.setdefault('format', config.get(__name__, 'format'))

    logger.debug("preparing containers")
    containers['input'] = Container(input_path)
    containers['output'] = Container(output_path, create=True)

    logger.debug("preparing file names")
    files['input'] = containers['input'].names()
    files['existing'] = containers['output'].names()
    files['stems'] = [splitext(files['input'][i])[0] for i in [0, -1]]
    files['output'] = f"{'-'.join(files['stems'])}.{kwargs['format']}"
    files['target'] = f"({containers['output']})[{files['output']}]"

    logger.debug("checking if %s already exists", files['output'])
    if files['output'] in files['existing']:
        logger.debug("skipping %s", files['output'])
        return

    logger.debug("loading data")
    data['input'] = []
    containers['iterator'] = containers['input'].iterator(files['input'])
    for i in range(len(files['input'])):
        logger.debug("loading %s", files['input'][i])
        series['input'] = next(containers['iterator'])
        data['input'].append(deserialize(series['input'], files['input'][i]))

    logger.debug("checking files")
    try:
        data['type'] = data['input'][0]['metadata']['type']
    except (KeyError, TypeError) as exc:
        raise RuntimeError("can not parse data from container") from exc
    if data['type'] == 'fourier-transform':
        process_transforms(containers, files, data)
    elif data['type'] == 'spatial-analysis':
        raise NotImplementedError()
    else:
        raise TypeError(f"can not merge data of type {data['type']}")


def process_transforms(containers, files, data):
    """Merge the Fourier transforms.

    Parameters
    ----------
    containers : dict of Container
        Input and output containers.
    files : dict of list
        Input and output files names.
    data : dict
        Input and output files data.

    """
    series = {}

    logger.debug("gathering coefficients")
    data['output'] = dict(data['input'][0])
    data['cos'] = []
    data['sin'] = []
    for i in range(len(data['input'])):
        if data['input'][i]['metadata']['type'] != 'fourier-transform':
            raise TypeError(f"{files['input'][i]} is not a Fourier transform")
        data['cos'].append(data['input'][i]['coefficients'].pop('cos_mean'))
        data['sin'].append(data['input'][i]['coefficients'].pop('sin_mean'))

    logger.debug("conputing mean and deviation")
    data['avg_cos'] = np.average(data['cos'], axis=0)
    data['avg_sin'] = np.average(data['sin'], axis=0)
    data['coefficient'] = 1/np.sqrt(len(files['input']))
    data['std_cos'] = np.std(data['cos'], axis=0) * data['coefficient']
    data['std_sin'] = np.std(data['sin'], axis=0) * data['coefficient']

    logger.info("generating %s", files['target'])
    data['output']['diffraction']['samples'] = len(files['input'])
    data['output']['coefficients']['cos_mean'] = data['avg_cos'].tolist()
    data['output']['coefficients']['sin_mean'] = data['avg_sin'].tolist()
    data['output']['coefficients']['cos_deviation'] = data['std_cos'].tolist()
    data['output']['coefficients']['sin_deviation'] = data['std_sin'].tolist()
    data['output']['metadata'] = {
        'type': 'fourier-transform',
        'date': datetime.now(timezone.utc).isoformat(timespec="seconds"),
        'version': version_tuple,
    }

    logger.debug("saving %s", files['output'])
    series['output'] = serialize(data['output'], files['output'])
    containers['output'].add(series['output'], files['output'])



# -*- coding: utf-8 -*-

"""Spatial analysis feature.

This module implements the functionality to compute the strain energy
and the interdislocation statistics of a group of dislocation samples.

"""

from datetime import datetime, timezone
from logging import getLogger
from os.path import splitext

from numlpa import config, kits
from numlpa._version import version_tuple
from numlpa.core.parallelism import get_rank, get_size
from numlpa.core.storage import Container, deserialize, serialize


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.set_defaults(
        func=main,
    )
    parser.add_argument(
        'input_path',
        type=str,
        help="path of the input container",
    )
    parser.add_argument(
        'output_path',
        type=str,
        help="path of the output container",
    )
    parser.add_argument(
        '-f',
        '--format',
        type=str,
        choices=('json', 'pyc'),
        help="format of the output files",
    )

    logger.debug("defining command-line subparsers")
    subparsers = parser.add_subparsers(
        dest='analyzer',
        help="name of the module used for analysis",
    )
    for name in kits.names('analyzers'):
        subparser = subparsers.add_parser(
            name=name,
            help=f'analyze with {name} analyzer',
        )
        module = kits.get('analyzers', name)
        module.setup(subparser)


def main(input_path, output_path, **kwargs):
    """Analize.

    Parameters
    ----------
    input_path : str
        Path of the input container.
    output_path : str
        Path of the output container.

    Keyword Arguments
    -----------------
    analyzer : str
        Name of the module used for analizing.
    format : str
        Format of the output files.

    """
    containers = {}
    files = {}
    modules = {}

    logger.debug("retrieving parameters")
    kwargs.setdefault('analyzer', config.get(__name__, 'analyzer'))
    kwargs.setdefault('format', config.get(__name__, 'format'))

    logger.debug("preparing containers")
    containers['input'] = Container(input_path)
    containers['output'] = Container(output_path, create=True)

    logger.debug("preparing file names")
    files['input'] = containers['input'].names()
    files['stems'] = [splitext(name)[0] for name in files['input']]
    files['output'] = [f"{stem}.{kwargs['format']}" for stem in files['stems']]
    files['existing'] = containers['output'].names()

    logger.debug("preparing submodules")
    modules['analyze'] = kits.get('analyzers', kwargs['analyzer'])

    logger.debug("generating files")
    static = {
        'containers': containers,
        'files': files,
        'kwargs': kwargs,
        'modules': modules,
    }
    for i in range(get_rank(), len(files['input']), get_size()):
        process(static, i)


def process(static, i):
    """Process a file.

    Parameters
    ----------
    static: dict of dict
        The data that does not vary from one file to another.
    i : int
        Index of the file to be generated.

    """
    data = {}
    file = {}
    passed = {}
    series = {}

    logger.debug("defining local variables for file index %d", i)
    file['input'] = static['files']['input'][i]
    file['output'] = static['files']['output'][i]
    file['target'] = f"({static['containers']['output']})[{file['output']}]"
    passed.update(static['kwargs'])

    logger.debug("checking if %s already exists", file['output'])
    if file['output'] in static['files']['existing']:
        logger.debug("skipping %s", file['output'])
        return

    logger.debug("loading %s", file['input'])
    series['input'] = static['containers']['input'].get(file['input'])
    data['input'] = deserialize(series['input'], file['input'])

    logger.debug("checking %s", file['input'])
    try:
        data['type'] = data['input']['metadata']['type']
    except (KeyError, TypeError) as exc:
        raise RuntimeError(f"can not parse data from {file['input']}") from exc
    if data['type'] != 'dislocation-sample':
        raise TypeError(f"invalid data type {data['type']}")

    logger.info("generating %s", file['target'])
    data['output'] = {
        'metadata': {
            'type': 'spatial-analysis',
            'date': datetime.now(timezone.utc).isoformat(timespec="seconds"),
            'version': version_tuple,
        },
        **static['modules']['analyze'].analyze(data['input'], **passed),
    }

    logger.debug("saving %s", file['output'])
    series['output'] = serialize(data['output'], file['output'])
    static['containers']['output'].add(series['output'], file['output'])



# -*- coding: utf-8 -*-

"""Wilkens model.

In 1970, M. Wilkens [1]_ proposed a model that has been simplified here
as follows:

.. math::

   \\ln\\left[ A(L) \\right] = - \\Lambda \\rho L^2 f\\left(
   \\frac{L}{R_e} \\right)

With:

.. math::

   \\Lambda = \\frac{\\pi}{2} g^2 b^2 C_g

.. math::

    f(\\eta) =
    \\begin{cases}
    - \\ln(\\eta) + \\frac{7}{4} - \\ln(2) + \\frac{512}{90 \\pi}
    \\eta^{-1} \\\\
    + \\frac{2}{\\pi} \\left( 1 - \\frac{1}{4}\\eta^{-2} \\right)
    \\int_0^\\eta \\frac{\\arcsin(V)}{V} dV \\\\
    - \\left( \\frac{769}{180} \\eta^{-1} + \\frac{41}{90} \\eta +
    \\frac{2}{90}\\eta^3 \\right) \\frac{\\sqrt{ 1 - \\eta^2 }}{\\pi}
    & \\forall \\eta \\leq 1 \\\\[3mm]
    \\frac{512}{90 \\pi} \\eta^{-1} - \\left( \\frac{11}{24} +
    \\frac{1}{4} \\ln(2) \\eta \\right) \\eta^{-2}
    & \\forall \\eta > 1
    \\end{cases}

* :math:`A` Fourier amplitude.

* :math:`L` Fourier variable (:math:`\\mathrm{m}`).

* :math:`\\vec{g}` diffraction vector (:math:`\\mathrm{m}^{-1}`).

* :math:`\\vec{b}` Burgers vector (:math:`\\mathrm{m}`).

* :math:`C_g` contrast factor.

* :math:`\\rho` dislocation density (:math:`\\mathrm{m}^{-2}`).

* :math:`R_e` radius (:math:`\\mathrm{m}`).

For this model, the parameters to be adjusted are :math:`\\rho` and
:math:`R_e`.

.. [1] M. Wilkens.
   Fundamental aspects of dislocation theory. Ed. by J. A. Simmons,
   R. de Wit, and R. Bullough. Vol. 2. U.S. National Bureau of
   Standards, 1970, pp. 1195-1221.

"""

from logging import getLogger

import numpy as np
import scipy.integrate


logger = getLogger(__name__)


def guide(transform):
    """Return a guide for the model parameters.

    Parameters
    ----------
    transform : dict
        Fourier transform data.

    Returns
    -------
    dict
        Parameter names, min, max and real values.

    """
    logger.debug("retrieving distribution parameters")
    density = transform['distribution']['density']
    distance = 1/np.sqrt(density)

    logger.debug("assembling parameter guide")
    data = {
        'names': ('density', 'radius'),
        'max': (density*100, distance/10),
        'min': (density/100, distance*1000),
        'real': (density, None),
    }
    return data


def model(transform, harmonic, limit=None):
    """Return the Fourier amplitude function.

    Parameters
    ----------
    transform : dict
        Fourier transform data.
    harmonic : int
        Harmonic of the diffraction vector.
    limit : int
        Index of the last Fourier variable.

    Returns
    -------
    Callable
        Function that takes parameter values and returns amplitudes.

    """
    data = {}

    logger.debug("defining constant quantities")
    data['L'] = np.array(transform['coefficients']['variable'])
    if limit:
        data['L'] = data['L'][:limit]
    data['b_uvw'] = np.array(transform['diffraction']['b_uvw'])
    data['g_hkl'] = np.array(transform['diffraction']['g_hkl'])
    data['g_vec'] = harmonic * data['g_hkl']/transform['diffraction']['cell']
    data['b_vec'] = data['b_uvw'] * transform['diffraction']['cell']/2
    data['g_len'] = np.linalg.norm(data['g_vec'])
    data['b_len'] = np.linalg.norm(data['b_vec'])
    data['C'] = transform['diffraction']['contrast']
    data['Lambda'] = np.pi/2 * data['g_len']**2 * data['b_len']**2 * data['C']
    data['cst1'] = data['Lambda'] * data['L']**2

    def amplitude(density, radius_e):
        if density < 0 or radius_e < 0:
            return np.inf
        return np.exp(-data['cst1']*density*vectorized_f(data['L']/radius_e))

    return amplitude


def wilkens_f(eta):
    """Return the result of the Wilkens intermediate function f.

    Parameters
    ----------
    eta : float
        Parameter of the Wilkens intermediate function f.

    Returns
    -------
    float
        Result of the Wilkens intermediate function f.

    """
    eta2 = eta**2

    def aux(val):
        if val == 0:
            return 1
        return np.arcsin(val)/val

    if eta < 1:
        result = (
            7/4 - np.log(eta) - np.log(2) + 512/90/np.pi/eta
            + 2/np.pi*(1-1/4/eta2)*scipy.integrate.quad(aux, 0, eta)[0]
            - 1/np.pi*(769/180/eta+41*eta/90+2*eta2*eta/90)*(1-eta2)**(1/2)
            - 1/np.pi*(11/12/eta2+7/2+eta2/3)*np.arcsin(eta) + eta2/6
        )
    else:
        result = 512/90/np.pi/eta - (11/24+np.log(2)*eta/4)/eta2
    return result


vectorized_f = np.vectorize(wilkens_f)



# -*- coding: utf-8 -*-

"""Spatial analyzers.

This package gathers the implementations of the spatial analyzers.
These compute the strain energy contained in a dislocation sample and
the distribution of dislocations around a dislocation. The strain
energy contained in the region of interest is given by:

.. math::

   E_T = \\sum_{i=1}^n E_{S, i} + \\frac{1}{2} \\sum_{i=1}^n
   \\sum_{\\substack{j=1 \\\\ j \\neq i}}^n E_{I, i, j}

With:

* :math:`E_T` total energy per unit length (:math:`\\mathrm{J.m}^{-1}`).

* :math:`E_{S, i}` self energy per unit length of the dislocation
  :math:`i` (:math:`\\mathrm{J.m}^{-1}`).

* :math:`E_{I, i, j}` interaction energy per unit length of :math:`i`
  and :math:`j` (:math:`\\mathrm{J.m}^{-1}`).

* :math:`n` number of dislocations in the region of interest.

The self energy of a dislocation and the interaction energy between two
dislocations are expressed respectively by: [1]_ [2]_

.. math::

   E_{S,i} = \\begin{cases}
   \\frac{G b^2}{4 \\pi} \\ln\\left( \\frac{r_{\\text{ROI}}}{r_0}
   \\right) & \\text{for screw dislocations} \\\\
   \\frac{G b^2}{4 \\pi (1-\\nu)} \\ln\\left( \\frac{r_{
   \\text{ROI}}}{r_0} \\right) & \\text{for edge dislocations}
   \\end{cases}

and

.. math::

   E_{I, i, j} = \\begin{cases}
   \\vec{b}_i \\cdot \\vec{b}_j \\frac{G}{2 \\pi} \\ln\\left(
   \\frac{r_{\\text{ROI}}}{d_{i, j}} \\right) &
   \\text{for screw dislocations} \\\\
   \\vec{b}_i \\cdot \\vec{b}_j \\frac{G}{2 \\pi (1-\\nu)}
   \\ln\\left( \\frac{r_{ \\text{ROI}}}{d_{i, j}} \\right) &
   \\text{for edge dislocations}
   \\end{cases}

With:

* :math:`G` shear modulus (:math:`\\mathrm{Pa}`).

* :math:`\\vec{b}` Burgers vector (:math:`\\mathrm{m}`).

* :math:`r_{\\text{ROI}}` size of the region of interest
  (:math:`\\mathrm{m}`).

* :math:`r_0` dislocation core radius (:math:`\\mathrm{m}`).

* :math:`d_{i,j}` distance between dislocations :math:`i` and :math:`j`
  (:math:`\\mathrm{m}`).

Finally we introduce the outer cut-off radius with the following
expression:

.. math::

   E_T = \\begin{cases}
   \\frac{G b^2 n}{2 \\pi} \\ln\\left(
   \\frac{r_{\\text{cut}}}{r_0} \\right) &
   \\text{for screw dislocations} \\\\
   \\frac{G b^2 n}{2 \\pi (1-\\nu)}
   \\ln\\left( \\frac{r_{\\text{cut}}}{r_0} \\right) &
   \\text{for edge dislocations}
   \\end{cases}

With:

* :math:`r_{\\text{cut}}` outer cut-off radius (:math:`\\mathrm{m}`).

.. [1] D. Hull and D. J. Bacon.
   Introduction to dislocations. Elsevier, 2011. ISBN:
   978-0-08-096672-4. DOI: 10.1016/C2009-0-64358-0.

.. [2] F. R. N. Nabarro.
   Theory of crystal dislocations. Oxford Clarendon Press, 1967.

"""



# -*- coding: utf-8 -*-

"""CPU diffractometer argument parser.

"""

from logging import getLogger


logger = getLogger(__package__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.add_argument(
        '--z_uvw',
        help="direction of the line vector (uvw)",
        type=int,
        nargs=3,
        metavar='int',
    )
    parser.add_argument(
        '--x_uvw',
        help="direction of the Fourier variable (uvw)",
        type=int,
        nargs=3,
        metavar='int',
    )
    parser.add_argument(
        '--b_uvw',
        help="direction of the Burgers vector (uvw)",
        type=int,
        nargs=3,
        metavar='int',
    )
    parser.add_argument(
        '--g_hkl',
        help="direction of the diffraction vector (hkl)",
        type=int,
        nargs=3,
        metavar='int',
    )
    parser.add_argument(
        '--cell',
        help="lattice constant (m)",
        type=float,
        metavar='float',
    )
    parser.add_argument(
        '--step',
        help="step size of the Fourier variable (m)",
        type=float,
        metavar='float',
    )
    parser.add_argument(
        '--range',
        help="absolute or relative limit of the Fourier variable (1|m)",
        type=float,
        metavar='float',
    )
    parser.add_argument(
        '--absolute',
        action='store_true',
        help="if true, the range parameter becomes an absolute value",
        default=False,
    )
    parser.add_argument(
        '--poisson',
        help="Poisson number",
        type=float,
        metavar='float',
    )
    parser.add_argument(
        '--harmonics',
        help="list of harmonics to be calculated",
        type=int,
        nargs='+',
        metavar='int',
    )
    parser.add_argument(
        '--replicate',
        help="number of replication of the region of interest",
        type=int,
        metavar='int',
    )
    parser.add_argument(
        '--points',
        help="number of random points",
        type=int,
        metavar='int',
    )
    parser.add_argument(
        '--processes',
        help="number of parallel processes",
        type=int,
        metavar='int',
    )
    parser.add_argument(
        '--check',
        action='store_const',
        const=True,
        help="display random points and replicated dislocations",
    )



# -*- coding: utf-8 -*-

"""Default optimizer.

This module implements a parallelized optimizer with random restart.

"""

from logging import getLogger
from multiprocessing import Pool
from os import cpu_count
from random import seed as set_seed, uniform

import numpy as np
from scipy import optimize

from numlpa import config


logger = getLogger(__name__)

shared = {}


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.add_argument(
        '--maxiter',
        type=int,
        help="maximum number of iterations",
        metavar='int',
    )
    parser.add_argument(
        '--method',
        type=str,
        help="name of the solver",
    )
    parser.add_argument(
        '--points',
        type=int,
        help="number of random initializations tested",
    )
    parser.add_argument(
        '-s',
        '--seed',
        type=int,
        help="random seed",
        metavar='int',
    )
    parser.add_argument(
        '--processes',
        type=int,
        help="number of parallel processes",
    )


def minimize(objective, guide, **kwargs):
    """Return the parameters minimizing the objective function.

    Parameters
    ----------
    objective : Callable
        Function to be minimized.
    guide : dict of list
        Parameter guide.

    Keyword Arguments
    -----------------
    maxiter : int
        Maximum number of iterations.
    method : str
        Name of the solver.
    points : int
        Number of random initializations tested.
    seed : int
        Random seed.
    processes : int
        Number of parallel processes.

    Returns
    -------
    parameters : list
        Optimal parameters values.

    """
    data = {}

    logger.debug("retrieving parameters")
    kwargs.setdefault('maxiter', config.getint(__name__, 'maxiter'))
    kwargs.setdefault('method', config.get(__name__, 'method'))
    kwargs.setdefault('points', config.getint(__name__, 'points'))
    kwargs.setdefault('seed', config.getint(__name__, 'seed'))
    kwargs.setdefault('processes', config.getint(__name__, 'processes'))
    if kwargs['processes'] < 1:
        kwargs['processes'] = cpu_count()

    logger.debug("defining initializations")
    set_seed(kwargs['seed'])
    data['inits'] = []
    for _ in range(kwargs['points']):
        init = []
        for i in range(len(guide['names'])):
            if guide['real'][i] is not None:
                init.append(guide['real'][i])
            else:
                init.append(uniform(guide['min'][i], guide['max'][i]))
        data['inits'].append(init)

    logger.debug("optimizing")
    share = {
        'fun': objective,
        'method': kwargs['method'],
        'options': {
            'maxiter': kwargs['maxiter'],
        }
    }
    np.seterr(all="ignore")
    with Pool(
        kwargs['processes'],
        initargs=(share,),
        initializer=initializer,
    ) as pool:
        data['result'] = pool.map(result, data['inits'])

    logger.debug("retrieving best parameters")
    data['rss'] = [r.fun for r in data['result']]
    if not any(r.success for r in data['result']):
        logger.warning("no attempt to adjust the model was successful")
    return data['result'][np.argmin(data['rss'])].x


def initializer(share):
    """Initialize a worker.

    Parameters
    ----------
    share : dict
        Shared data to be passed to the worker.

    """
    shared.update(share)


def result(initialization):
    """Return the optimization result for the given initialization.

    Parameters
    ----------
    initialization : list
        Parameter initialization.

    Returns
    -------
    OptimizeResult
        Optimization result.

    """
    return optimize.minimize(x0=initialization, **shared)



# -*- coding: utf-8 -*-

"""Affine data representations.

This module implements the figure generators allowing to show the
asymptotic affine behavior of the Fourier transform amplitude after
the following transformation:

.. math::

   A(L) \\mapsto \\frac{ln(A(L))}{L^2}

"""

from json import loads
from logging import getLogger

import matplotlib.pyplot as plt
import numpy as np

from numlpa import config


logger = getLogger(__package__)

TEX_G = r"\vec{g}"
TEX_U = r"\vec{u}"
TEX_R = r"\vec{r}"
TEX_L = r"\vec{L}"

CONVERSION = ['']


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.add_argument(
        '--figsize',
        help="width and heigth of the figure (inches)",
        type=float,
        nargs=2,
        metavar='float',
    )
    parser.add_argument(
        '--title',
        help="title of the figure",
        type=str,
        metavar='str',
    )
    legend = parser.add_mutually_exclusive_group()
    legend.add_argument(
        '--legend',
        action='store_const',
        const=True,
        help="display legends",
        dest='legend'
    )
    legend.add_argument(
        '--no-legend',
        action='store_const',
        const=False,
        help="hide legends",
        dest='legend'
    )
    parser.add_argument(
        '--length',
        help="length unit",
        type=str,
        choices=('m', 'nm'),
    )


def retrieve_parameters(kwargs):
    """Load the default value of the missing parameters.

    Parameters
    ----------
    kwargs : dict
        Parameters passed to a function of the module.

    """
    logger.debug("retrieving parameters")
    kwargs.setdefault('figsize', loads(config.get(__name__, 'figsize')))
    kwargs.setdefault('title', config.get(__name__, 'title'))
    kwargs.setdefault('legend', config.getboolean(__name__, 'legend'))
    kwargs.setdefault('length', config.get(__name__, 'length'))


def log_x2(amplitude, variable):
    """Return the transformed amplitude.

    Parameters
    ----------
    amplitude : list or np.array
        Fourier amplitude values.
    variable : list or np.array
        Fourier variable values.

    Returns
    -------
    np.array
        Value of 'ln(amplitude)/variable**2'.

    """
    amplitude = np.array(amplitude)
    variable = np.array(variable)
    amplitude = np.where(amplitude > 0, amplitude, np.nan)
    return np.log(amplitude)/variable**2


def transform(data, **kwargs):
    """Return a matplotlib figure illustrating a Fourier transform.

    Parameters
    ----------
    data : dict
        Fourier transform data.

    Keyword Arguments
    -----------------
    figsize : tuple
        Width and heigth of the figure (inches).
    title : str
        Title of the figure.
    legend : bool
        Display legends.
    length : str
        Length unit ('m' or 'nm').

    Returns
    -------
    Figure
        Matplotlib figure.

    """
    figure = {}
    retrieve_parameters(kwargs)

    logger.debug("initializing figure")
    figure['fig'], figure['ax'] = plt.subplots(figsize=kwargs['figsize'])
    figure['ax'].grid()
    figure['ax'].set_xscale('log')

    logger.debug("setting scale coefficient for unit conversion")
    figure['scale'] = 1
    if kwargs['length'] == 'nm':
        figure['scale'] = 1e9

    logger.debug("plotting profiles")
    figure['variable'] = np.array(data['coefficients']['variable'])
    figure['cos'] = {
        'avg': np.array(data['coefficients']['cos_mean']),
        'std': np.array(data['coefficients']['cos_deviation']),
    }
    figure['fill'] = {
        'min': figure['cos']['avg'] - figure['cos']['std'],
        'max': figure['cos']['avg'] + figure['cos']['std'],
    }
    for i, j in enumerate(data['coefficients']['harmonic']):
        figure['ax'].plot(
            figure['scale']*figure['variable'],
            log_x2(figure['cos']['avg'][i], figure['variable']),
            label=(
                fr"$\ln \left("
                fr"\left\langle \cos \left( 2 \pi {TEX_G}_{j} \cdot "
                fr"\left[ {TEX_U} \left( {TEX_R} + {TEX_L} \right) "
                fr"- {TEX_U} \left( {TEX_R} \right) \right] \right) "
                fr"\right\rangle"
                fr"\right) / L^2$"
            )
        )
        figure['ax'].fill_between(
            figure['scale']*figure['variable'],
            log_x2(figure['fill']['min'][i], figure['variable']),
            log_x2(figure['fill']['max'][i], figure['variable']),
            alpha=0.3,
        )

    logger.debug("displaying text")
    if kwargs['legend']:
        figure['ax'].legend()
    if kwargs['title'] != "":
        figure['ax'].set_title(kwargs['title'])
    figure['ax'].set_xlabel(f"$L$ ({kwargs['length']})")

    return figure['fig']


def adjustment(data, **kwargs):
    """Return a matplotlib figure illustrating a Fourier transform.

    Parameters
    ----------
    data : dict
        Model adjustment data.

    Keyword Arguments
    -----------------
    figsize : tuple
        Width and heigth of the figure (inches).
    title : str
        Title of the figure.
    legend : bool
        Display legends.
    length : str
        Length unit ('m' or 'nm').

    Returns
    -------
    Figure
        Matplotlib figure.

    """
    figure = {}
    retrieve_parameters(kwargs)

    logger.debug("initializing figure")
    figure['fig'], figure['ax'] = plt.subplots(figsize=kwargs['figsize'])
    figure['ax'].grid()
    figure['ax'].set_xscale('log')

    logger.debug("setting scale coefficient for unit conversion")
    figure['scale'] = 1
    if kwargs['length'] == 'nm':
        figure['scale'] = 1e9

    logger.debug("plotting model and simulated profile")
    figure['ax'].plot(
        np.array(data['profile']['variable'])*figure['scale'],
        log_x2(data['profile']['amplitude'], data['profile']['variable']),
        label='profile',
    )
    figure['ax'].plot(
        np.array(data['model']['variable'])*figure['scale'],
        log_x2(data['model']['amplitude'], data['model']['variable']),
        label='model',
    )

    logger.debug("displaying text")
    if kwargs['legend']:
        figure['ax'].legend()
    if kwargs['title'] != "":
        figure['ax'].set_title(kwargs['title'])
    figure['ax'].set_xlabel(f"$L$ ({kwargs['length']})")

    return figure['fig']



# -*- coding: utf-8 -*-

"""Upgradeable collections of features.

This package groups by category the different implementations or
alternatives proposed to the user to perform a given task.

"""

from importlib import import_module
from os.path import join
from pkgutil import iter_modules


def names(subpackage):
    """Return the names of the modules contained in the subpackage.

    Parameters
    ----------
    subpackage : str
        Name of the subpackage.

    Returns
    -------
    list of str
        Names of the modules in the subpackage.

    """
    paths = [join(path, subpackage) for path in __path__]
    return [module.name for module in iter_modules(paths)]


def get(subpackage, module):
    """Import and return a module contained in a subpackage.

    Parameters
    ----------
    subpackage : str
        Name of the subpackage.
    module : str
        Name of the submodule.

    Returns
    -------
    ModuleType
        Module from the subpackage.

    """
    return import_module(f'{__name__}.{subpackage}.{module}')



# -*- coding: utf-8 -*-

"""CPU analyzer.

"""

from logging import getLogger
from multiprocessing import Pool, RawArray
from os import cpu_count
from time import time

import numpy as np

from numlpa import config
from . import interaction


logger = getLogger(__package__)

shared = {}


def analyze(sample, **kwargs):
    """Return the spatial analysis.

    Parameters
    ----------
    sample : dict
        Dislocation sample data.

    Keyword Arguments
    -----------------
    type : str
        Dislocation type ('screw' or 'edge').
    r_0 : float
        Core radius (m).
    r_roi : float
        Size of the region of interest (m).
    r_max : float
        Maximum radius value (m).
    steps : int
        Number of subdivision on the absice axis.
    b_len : float
        Burgers vector length (m).
    poisson : float
        Poisson number.
    shear : float
        Shear modulus (Pa).
    processes : int
        Number of parallel processes.

    Returns
    -------
    dict
        Spatial analyzis data.

    """
    settings = {}
    share = {}
    times = {}
    result = {}

    check_parameters(kwargs)
    prepare_settings(sample, kwargs, settings)
    prepare_shared_data(sample, settings, share)
    run_analysis(settings, share, result, times)
    return assemble_data(sample, settings, kwargs, result, times)


def check_parameters(kwargs):
    """Set the default configuration value to missing parameters.

    Parameters
    ----------
    kwargs : dict
        Keyword arguments passed to the main function.

    """
    logger.debug("retrieving parameters")
    kwargs.setdefault('type', config.get(__package__, 'type'))
    kwargs.setdefault('r_0', config.getfloat(__package__, 'r_0'))
    kwargs.setdefault('r_roi', config.getfloat(__package__, 'r_roi'))
    kwargs.setdefault('r_max', config.getfloat(__package__, 'r_max'))
    kwargs.setdefault('steps', config.getint(__package__, 'steps'))
    kwargs.setdefault('b_len', config.getfloat(__package__, 'b_len'))
    kwargs.setdefault('poisson', config.getfloat(__package__, 'poisson'))
    kwargs.setdefault('shear', config.getfloat(__package__, 'shear'))
    kwargs.setdefault('processes', config.getint(__package__, 'processes'))
    if kwargs['processes'] < 1:
        kwargs['processes'] = cpu_count()


def prepare_settings(sample, kwargs, settings):
    """Prepare the diffraction variables.

    Parameters
    ----------
    sample : dict
        Dislocation sample data.
    kwargs : dict
        Keyword arguments passed to the main function.
    settings : dict
        Diffraction variables.

    """
    logger.debug("adding some of the parameters passed to the main function")
    settings['r_0'] = kwargs['r_0']
    settings['r_roi'] = kwargs['r_roi']
    settings['steps'] = kwargs['steps']
    settings['b_len'] = kwargs['b_len']
    settings['poisson'] = kwargs['poisson']
    settings['shear'] = kwargs['shear']
    settings['processes'] = kwargs['processes']

    logger.debug("preparing energy parameters")
    settings['n_all'] = len(sample['dislocations']['senses'])
    settings['indexes'] = range(settings['n_all'])

    logger.debug("identifying region type")
    if sample['region']['type'] == 'square':
        settings['r_max'] = 2 * sample['region']['side']
        if settings['r_roi'] == 0:
            settings['r_roi'] = sample['region']['side']
    elif sample['region']['type'] == 'disk':
        settings['r_max'] = 2 * sample['region']['radius']
        if settings['r_roi'] == 0:
            settings['r_roi'] = sample['region']['radius']
    else:
        raise NotImplementedError("unknown region type")
    settings['step'] = settings['r_max'] / kwargs['steps']

    logger.debug("identifying dislocation type")
    if kwargs['type'] == 'edge':
        settings['factor'] = 1 / (1-settings['poisson'])
    elif kwargs['type'] == 'screw':
        settings['factor'] = 1
    else:
        raise NotImplementedError("mixed dislocations")


def prepare_shared_data(sample, settings, share):
    """Initialize and fill the shared containers.

    Parameters
    ----------
    sample : dict
        Dislocation sample data.
    settings : dict
        Diffraction variables.
    share : dict
        Shared data.

    """
    wrap = {}

    logger.debug("declaring shared data")
    share['x'] = RawArray('d', settings['n_all'])
    share['y'] = RawArray('d', settings['n_all'])
    share['s'] = RawArray('b', settings['n_all'])
    share['steps'] = settings['steps']
    share['step'] = settings['step']

    logger.debug("wrapping shared data")
    wrap['x'] = np.frombuffer(share['x'], dtype=np.float64)
    wrap['y'] = np.frombuffer(share['y'], dtype=np.float64)
    wrap['s'] = np.frombuffer(share['s'], dtype=np.int8)

    logger.debug("defining shared data")
    wrap['x'][:] = sample['dislocations']['positions'][0]
    wrap['y'][:] = sample['dislocations']['positions'][1]
    wrap['s'][:] = sample['dislocations']['senses']


def run_analysis(settings, share, result, times):
    """Run the simulation with Monte Carlo method.

    Parameters
    ----------
    settings : dict
        Diffraction variables.
    share : dict
        Shared data.
    results : dict
        Diffraction results.
    times : dict
        Time measurements.

    """
    logger.debug("computing self energy")
    result['self_energy'] = (
        settings['n_all'] *
        settings['shear'] *
        settings['b_len']**2 /
        (4 * np.pi) *
        settings['factor'] *
        np.log(settings['r_roi']/settings['r_0'])
    )

    logger.debug("computing interactions")
    times['0'] = time()
    with Pool(
        settings['processes'],
        initializer=initializer,
        initargs=(share,),
    ) as pool:
        result['raw'] = pool.map(interaction_data, settings['indexes'])
    times['1'] = time()

    logger.debug("processing output")
    result['interaction_energy'] = 0
    result['count_mm'] = np.zeros(settings['steps'], dtype=np.int64)
    result['count_mp'] = np.zeros(settings['steps'], dtype=np.int64)
    result['count_pm'] = np.zeros(settings['steps'], dtype=np.int64)
    result['count_pp'] = np.zeros(settings['steps'], dtype=np.int64)
    for i in settings['indexes']:
        term, count_equal, count_opposed = result['raw'][i]
        result['interaction_energy'] += term
        if share['s'][i] > 0:
            result['count_pp'] += count_equal
            result['count_pm'] += count_opposed
        else:
            result['count_mm'] += count_equal
            result['count_mp'] += count_opposed
    result['interaction_energy'] += (
        settings['n_all'] *
        (settings['n_all']-1) *
        np.log(settings['r_roi'])
    )
    result['interaction_energy'] *= (
        1/2 *
        settings['shear'] *
        settings['b_len']**2 /
        (4 * np.pi) *
        settings['factor']
    )

    logger.debug("computing total energy")
    result['total_energy'] = result['self_energy']
    result['total_energy'] += result['interaction_energy']

    logger.debug("computing outer cut-off radius")
    result['r_cut'] = settings['r_0'] * np.exp(
        result['total_energy'] *
        (2 * np.pi) /
        settings['n_all'] /
        settings['shear'] /
        settings['b_len']**2 /
        settings['factor']
    )


def initializer(share):
    """Initialize a worker.

    Parameters
    ----------
    share : dict
        Shared data to be passed to the worker.

    """
    shared.update(share)


def assemble_data(sample, settings, kwargs, result, times):
    """Return the Fourier transform data.

    Parameters
    ----------
    sample : dict
        Dislocation sample data.
    settings : dict
        Diffraction variables.
    kwargs : dict
        Keyword arguments passed to the main function.
    results : dict
        Diffraction results.
    times : dict
        Time measurements.

    Returns
    -------
    dict
        Fourier transform data.

    """
    logger.debug("assembling data")
    data = {
        'distribution': sample['distribution'],
        'region': sample['region'],
        'analysis': {
            'module': __package__,
            'type': kwargs['type'],
            'r_0': settings['r_0'],
            'r_roi': settings['r_roi'],
            'r_max': settings['r_max'],
            'steps': settings['steps'],
            'poisson': settings['poisson'],
            'shear': settings['shear'],
            'r_cut': result['r_cut'],
            'self_energy': result['self_energy'],
            'interaction_energy': result['interaction_energy'],
            'total_energy': result['total_energy'],
            'variable': [settings['step']*i for i in range(kwargs['steps'])],
            'count_mm': result['count_mm'].tolist(),
            'count_mp': result['count_mp'].tolist(),
            'count_pm': result['count_pm'].tolist(),
            'count_pp': result['count_pp'].tolist(),
            'samples': 1,
            'duration': times['1'] - times['0'],
            'processes': kwargs['processes'],
        }
    }
    return data


def interaction_data(index):
    """Return the interaction data for one dislocation.

    This function computes the interaction energy term and the
    dislocation distribution around a dislocation using the interaction
    extension module.

    Parameters
    ----------
    index : int
        Index of the dislocation.

    Returns
    -------
    tuple
        Interaction data for screw dislocations.

    """
    count_equal = RawArray('Q', shared['steps'])
    count_opposed = RawArray('Q', shared['steps'])
    np.frombuffer(count_equal, dtype=np.uint64)[:] = 0
    np.frombuffer(count_opposed, dtype=np.uint64)[:] = 0
    term = interaction.sum_term(
        index,
        shared['x'],
        shared['y'],
        shared['s'],
        shared['step'],
        count_equal,
        count_opposed,
    )
    raw = (
        term,
        np.array(count_equal, dtype=np.int64),
        np.array(count_opposed, dtype=np.int64),
    )
    return raw



# -*- coding: utf-8 -*-

"""LPA theoretical models.

This package gathers the implementations of available LPA models for
the prediction of microstructure parameters from the Fourier transform
of the diffracted intensity.

"""



# -*- coding: utf-8 -*-

"""Model adjustment feature.

This module implements the functionality to determine the parameter
values predicted by the models by fitting the latter to the diffraction
simulation results.

"""

from datetime import datetime, timezone
from logging import getLogger

import numpy as np

from numlpa import config, kits
from numlpa._version import version_tuple
from numlpa.core.parallelism import get_rank, get_size
from numlpa.core.storage import Container, deserialize, serialize


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.set_defaults(
        func=main,
    )
    parser.add_argument(
        'input_path',
        type=str,
        help="path of the input container",
    )
    parser.add_argument(
        'output_path',
        type=str,
        help="path of the output container",
    )
    parser.add_argument(
        '-f',
        '--format',
        type=str,
        choices=('json', 'pyc'),
        help="format of the output files",
    )
    parser.add_argument(
        '--model',
        type=str,
        choices=kits.names('models'),
        help="name of the model to be fitted",
    )
    parser.add_argument(
        '--restrict',
        type=str,
        choices=kits.names('restrictions'),
        help="name of the restriction for the variable interval",
    )
    parser.add_argument(
        '--part',
        type=str,
        choices=('real', 'module'),
        help="part of the Fourier transform to be fitted",
    )
    parser.add_argument(
        '--harmonic',
        type=int,
        metavar='int',
        help="harmonic of the Fourier transform to be fitted",
    )

    logger.debug("defining command-line subparsers")
    subparsers = parser.add_subparsers(
        dest='optimizer',
        help="name of the module used for error minimization",
    )
    for name in kits.names('optimizers'):
        subparser = subparsers.add_parser(
            name=name,
            help=f'minimize with {name} optimizer',
        )
        module = kits.get('optimizers', name)
        module.setup(subparser)


def main(input_path, output_path, **kwargs):
    """Adjust a model.

    Parameters
    ----------
    input_path : str
        Path of the input container.
    output_path : str
        Path of the output container.

    Keyword Arguments
    -----------------
    input_path : str
        Path of the input container.
    output_path : str
        Path of the output container.
    format : str
        Format of the output files.
    model : str
        Name of the model to be fitted.
    restrict : str
        Name of the restriction for the variable interval.
    part : str
        Part of the Fourier transform to be fitted ('real', 'module').
    harmonic : int
        Harmonic of the Fourier transform to be fitted.
    optimizer : str
        Name of the module used for error minimization.

    """
    containers = {}
    data = {}
    files = {}
    modules = {}
    series = {}

    logger.debug("retrieving arguments")
    kwargs.setdefault('format', config.get(__name__, 'format'))
    kwargs.setdefault('model', config.get(__name__, 'model'))
    kwargs.setdefault('restrict', config.get(__name__, 'restrict'))
    kwargs.setdefault('part', config.get(__name__, 'part'))
    kwargs.setdefault('harmonic', config.getint(__name__, 'harmonic'))
    kwargs.setdefault('optimizer', config.get(__name__, 'optimizer'))

    logger.debug("preparing containers")
    containers['input'] = Container(input_path)
    containers['output'] = Container(output_path, create=True)

    logger.debug("preparing submodules")
    modules['opti'] = kits.get('optimizers', kwargs['optimizer'])
    modules['model'] = kits.get('models', kwargs['model'])

    logger.debug("loading data")
    files['input'] = containers['input'].names()
    if len(files['input']) > 1:
        raise RuntimeError("multiple files in input container, merge them")
    series['input'] = containers['input'].get(files['input'][0])
    data['input'] = deserialize(series['input'], files['input'][0])
    data['harmonics'] = data['input']['coefficients']['harmonic']
    data['i_harmonic'] = data['harmonics'].index(kwargs['harmonic'])
    data['profile'] = {
        'var': data['input']['coefficients']['variable'],
        'cos': data['input']['coefficients']['cos_mean'][data['i_harmonic']],
        'sin': data['input']['coefficients']['sin_mean'][data['i_harmonic']],
    }
    data['guide'] = modules['model'].guide(data['input'])

    logger.debug("pre-processing profile")
    if kwargs['part'] == 'real':
        data['profile']['amp'] = data['profile']['cos']
    elif kwargs['part'] == 'module':
        data['profile']['amp'] = np.sqrt(
            data['profile']['cos']**2 + data['profile']['sin']**2
        )
    else:
        raise ValueError(f"unknown part '{kwargs['part']}'")
    data['limits'] = {}
    for name in kits.names('restrictions'):
        modules['restrict'] = kits.get('restrictions', name)
        data['limits'][name] = modules['restrict'].limit(
            data['profile']['var'],
            data['profile']['amp'],
        )
    data['imaxmax'] = data['limits'][kwargs['restrict']]
    data['imax'] = list(range(2, data['imaxmax']+1))

    logger.debug("preparing file names")
    width = len(str(data['imaxmax']))
    files['stem'] = [f'0-{str(i).zfill(width)}' for i in data['imax']]
    files['output'] = [f"{stem}.{kwargs['format']}" for stem in files['stem']]
    files['existing'] = containers['output'].names()

    logger.debug("generating files")
    static = {
        'containers': containers,
        'files': files,
        'kwargs': kwargs,
        'modules': modules,
        'data': data,
    }
    for i in range(get_rank(), len(data['imax']), get_size()):
        process(static, i)


def process(static, i):
    """Process a file.

    Parameters
    ----------
    static: dict of dict
        The data that does not vary from one file to another.
    i : int
        Index of the file to be generated.

    """
    file = {}
    fit = {}
    passed = {}
    series = {}

    logger.debug("defining local variables for file index %d", i)
    file['output'] = static['files']['output'][i]
    file['target'] = f"({static['containers']['output']})[{file['output']}]"
    passed.update(static['kwargs'])

    logger.debug("checking if %s already exists", file['output'])
    if file['output'] in static['files']['existing']:
        logger.debug("skipping %s", file['output'])
        return

    logger.info("generating %s", file['target'])
    fit['imax'] = static['data']['imax'][i]
    fit['amplitude'] = np.array(static['data']['profile']['amp'][:fit['imax']])
    fit['variable'] = np.array(static['data']['profile']['var'][:fit['imax']])
    fit['function'] = static['modules']['model'].model(
        static['data']['input'],
        static['kwargs']['harmonic'],
        fit['imax'],
    )

    def objective(parameters):
        return rss(fit['amplitude'], fit['function'], parameters)

    logger.debug("fitting over range [0, %d]", fit['imax'])
    fit['parameters'] = static['modules']['opti'].minimize(
        objective,
        static['data']['guide'],
        **passed,
    )

    logger.debug("computing relative values")
    fit['relative'] = [None for _ in static['data']['guide']['names']]
    for j in range(len(static['data']['guide']['names'])):
        real = static['data']['guide']['real'][j]
        if real is not None:
            fit['relative'][j] = fit['parameters'][j] / real

    logger.debug("assembling data")
    fit['output'] = {
        'metadata': {
            'type': 'model-adjustment',
            'date': datetime.now(timezone.utc).isoformat(timespec="seconds"),
            'version': version_tuple,
        },
        'distribution': static['data']['input']['distribution'],
        'region': static['data']['input']['region'],
        'diffraction': static['data']['input']['diffraction'],
        'profile': {
            'harmonic': static['kwargs']['harmonic'],
            'part': static['kwargs']['part'],
            'variable': list(static['data']['profile']['var']),
            'amplitude': list(static['data']['profile']['amp']),
        },
        'limits': static['data']['limits'],
        'model': {
            'module': static['modules']['model'].__name__,
            'variable': list(fit['variable']),
            'amplitude': list(fit['function'](*fit['parameters'])),
        },
        'optimization': {
            'module': static['modules']['opti'].__name__,
            'restriction': static['kwargs']['restrict'],
            'rss': rss(fit['amplitude'], fit['function'], fit['parameters']),
        },
        'parameters': {
            'names': static['data']['guide']['names'],
            'values': list(fit['parameters']),
            'real': static['data']['guide']['real'],
            'relative': fit['relative'],
        },
    }
    series['output'] = serialize(fit['output'], file['output'])
    static['containers']['output'].add(series['output'], file['output'])


def rss(amplitude, function, parameters):
    """Return the residual sum of squares.

    Parameters
    ----------
    amplitude : np.array
        Value of the true amplitude.
    function : Callable
        Function returning the amplitude given by the model.
    parameters : np.array
        Value of the function parameters.

    Returns
    -------
    float
        Residual sum of squares
    """
    return np.sum((function(*parameters) - amplitude)**2)



# -*- coding: utf-8 -*-

"""Numerical diffractometers.

This package gathers the implementations of the X-ray diffraction
simulators. These return the amplitude of the Fourier transform of
the diffracted intensity. According to the kinematical theory of
X-ray scattering [1]_, the Fourier transform is calculated from
the displacement field of the defects:

.. math::

   A(L) = \\frac{1}{V} \\int_V \\exp\\left(2 \\pi i \\vec{g} \\cdot
   \\left[\\vec{u}(\\vec{r}+\\vec{L}) - \\vec{u}(\\vec{r})\\right]
   \\right) d\\vec{r}

With:

* :math:`A` Fourier transform amplitude.

* :math:`\\vec{L}` Fourier variable vector, perpendicular to the
  diffracting lattice planes (:math:`\\mathrm{m}`).

* :math:`V` volume of the region of interest (:math:`\\mathrm{m}^{3}`).

* :math:`\\vec{g}` diffraction vector (:math:`\\mathrm{m}^{-1}`).

* :math:`\\vec{u}` displacement field (:math:`\\mathrm{m}`).

The resulting displacement field at a point is obtained by summing
the contributions of each dislocation. The displacement field induced
by a single screw or edge dislocation is given respectively by: [2]_

.. math::

   \\vec{u}_{\\text{screw}} = \\left( \\begin{array}{c}
   0 \\\\
   0 \\\\
   \\frac{b}{2 \\pi} \\arctan\\left(\\frac{y}{x}\\right)
   \\end{array} \\right)

and

.. math::

   \\vec{u}_{\\text{edge}} = \\left( \\begin{array}{c}
   \\frac{b}{2 \\pi} \\left( \\arctan\\left(\\frac{y}{x}\\right) +
   \\frac{xy}{2(1-\\nu)(x^2+y^2)} \\right) \\\\
   - \\frac{b}{8 \\pi (1 - \\nu)} \\left( (1-2\\nu) \\ln\\left(
   \\frac{x^2 + y^2}{b^2} \\right) - \\frac{2 y^2}{x^2 + y^2}
   \\right) \\\\
   0
   \\end{array} \\right)

With:

* :math:`b` Burgers vector length (:math:`\\mathrm{m}`).

* :math:`x` dislocation x-coordinate with respect to the observation
  position (:math:`\\mathrm{m}`).

* :math:`y` dislocation y-coordinate with respect to the observation
  position (:math:`\\mathrm{m}`).

* :math:`\\nu` Poisson's number.

.. [1] B. E. Warren.
   X-Ray diffraction. Dover Publications, 1990. ISBN: 978-0-48-666317-3.

.. [2] F. R. N. Nabarro.
   Theory of crystal dislocations. Oxford Clarendon Press, 1967.

"""



# -*- coding: utf-8 -*-

"""CPU diffractometer implementation.

"""

from json import loads
from logging import getLogger
from math import sqrt, isclose
from multiprocessing import Pool, RawArray
from os import cpu_count
from time import time

import matplotlib.pyplot as plt
import numpy as np

from numlpa import config
from . import displacement


logger = getLogger(__package__)

shared = {}


def diffract(sample, **kwargs):
    """Return the Fourier transform.

    Parameters
    ----------
    sample : dict
        Dislocation sample data.

    Keyword Arguments
    -----------------
    z_uvw : tuple of int
        Direction of the line vector (uvw).
    x_uvw : tuple of int
        Direction of the Fourier variable (uvw).
    b_uvw : tuple of int
        Direction of the Burgers vector (uvw).
    g_hkl : tuple of int
        Direction of the diffraction vector (hkl).
    cell : float
        Lattice constant (m).
    step : float
        Step size of the Fourier variable (m).
    range : float
        Absolute or relative limit of the Fourier variable (1|m).
    absolute : bool
        If true, the range parameter becomes an absolute value.
    poisson : float
        Poisson number.
    harmonics : list of int
        List of harmonics to be calculated.
    replicate : int
        Number of replication of the region of interest.
    points : int
        Number of random points.
    processes : int
        Number of parallel processes.
    check : bool
        Display random points and replicated dislocations.

    Returns
    -------
    dict
        Fourier transform data.

    """
    settings = {}
    share = {}
    times = {}
    result = {}

    check_parameters(kwargs)
    prepare_settings(sample, kwargs, settings)
    prepare_shared_data(sample, settings, share)
    user_check(settings, share, kwargs)
    run_monte_carlo(settings, share, result, times)
    return assemble_data(sample, settings, kwargs, result, times)


def check_parameters(kwargs):
    """Set the default configuration value to missing parameters.

    Parameters
    ----------
    kwargs : dict
        Keyword arguments passed to the main function.

    """
    logger.debug("retrieving parameters")
    kwargs.setdefault('z_uvw', loads(config.get(__package__, 'z_uvw')))
    kwargs.setdefault('x_uvw', loads(config.get(__package__, 'x_uvw')))
    kwargs.setdefault('b_uvw', loads(config.get(__package__, 'b_uvw')))
    kwargs.setdefault('g_hkl', loads(config.get(__package__, 'g_hkl')))
    kwargs.setdefault('cell', config.getfloat(__package__, 'cell'))
    kwargs.setdefault('step', config.getfloat(__package__, 'step'))
    kwargs.setdefault('range', config.getfloat(__package__, 'range'))
    kwargs.setdefault('absolute', config.getboolean(__package__, 'absolute'))
    kwargs.setdefault('points', config.getint(__package__, 'points'))
    kwargs.setdefault('poisson', config.getfloat(__package__, 'poisson'))
    kwargs.setdefault('harmonics', loads(config.get(__package__, 'harmonics')))
    kwargs.setdefault('processes', config.getint(__package__, 'processes'))
    kwargs.setdefault('replicate', config.getint(__package__, 'replicate'))
    kwargs.setdefault('check', config.getboolean(__package__, 'check'))
    if kwargs['processes'] < 1:
        kwargs['processes'] = cpu_count()


def unit_vector(vector):
    """Return the normalized vector.

    Parameters
    ----------
    vector : np.ndarray or list or tuple
        Vector to be normalized.

    Returns
    -------
    np.ndarray
        Normalized vector.

    """
    return np.array(vector) / np.linalg.norm(vector)


def prepare_settings(sample, kwargs, settings):
    """Prepare the diffraction variables.

    Parameters
    ----------
    sample : dict
        Dislocation sample data.
    kwargs : dict
        Keyword arguments passed to the main function.
    settings : dict
        Diffraction variables.

    """
    logger.debug("calculating the maximum value of the Fourier variable")
    settings['maximum'] = kwargs['range']
    if not kwargs['absolute']:
        settings['maximum'] /= sqrt(sample['distribution']['density'])
    settings['steps'] = int(settings['maximum']/kwargs['step'])

    logger.debug("preparing the change of reference frame")
    settings['z_uni_rf2'] = unit_vector(kwargs['z_uvw'])
    settings['x_uni_rf2'] = unit_vector(kwargs['x_uvw'])
    settings['y_uni_rf2'] = np.cross(
        settings['z_uni_rf2'],
        settings['x_uni_rf2'],
    )
    settings['transition'] = np.array((
        settings['x_uni_rf2'],
        settings['y_uni_rf2'],
        settings['z_uni_rf2'],
    ))

    logger.debug("defining vectors")
    settings['b_uni_rf2'] = unit_vector(kwargs['b_uvw'])
    settings['g_uni_rf2'] = unit_vector(kwargs['g_hkl'])
    settings['b_vec_rf2'] = np.array(kwargs['b_uvw']) * kwargs['cell']/2
    settings['g_vec_rf2'] = np.array(kwargs['g_hkl']) / kwargs['cell']
    settings['g_vec_rf1'] = settings['transition'].dot(settings['g_vec_rf2'])
    settings['g_uni_rf1'] = settings['transition'].dot(settings['g_uni_rf2'])
    settings['s_vec_rf1'] = kwargs['step'] * settings['g_uni_rf1']
    settings['b_len'] = np.linalg.norm(settings['b_vec_rf2'])

    logger.debug("calculating angles")
    settings['abs_ang_z_b'] = np.abs(np.arccos(np.dot(
        settings['z_uni_rf2'],
        settings['b_uni_rf2'],
    )))
    settings['abs_ang_z_g'] = np.abs(np.arccos(np.dot(
        settings['z_uni_rf2'],
        settings['g_uni_rf2'],
    )))

    logger.debug("defining the number of dislocations")
    settings['n_roi'] = len(sample['dislocations']['senses'])
    settings['n_all'] = settings['n_roi'] * (2*kwargs['replicate'] + 1)**2

    logger.debug("drawing random points")
    settings['points'] = random_points(sample, kwargs['points'])

    logger.debug("adding some of the parameters passed to the main function")
    settings['replicate'] = kwargs['replicate']
    settings['poisson'] = kwargs['poisson']
    settings['harmonics'] = kwargs['harmonics']
    settings['step'] = kwargs['step']
    settings['processes'] = kwargs['processes']

    logger.debug("identifying dislocation type")
    if isclose(settings['abs_ang_z_b'], np.pi/2, abs_tol=1e-4):
        settings['type'] = 'edge'
        settings['function'] = phases_edge
        settings['contrast'] = contrast_factor_edge(settings)
    elif isclose(settings['abs_ang_z_b'], 0, abs_tol=1e-4):
        settings['type'] = 'screw'
        settings['function'] = phases_screw
        settings['contrast'] = contrast_factor_screw(settings)
    else:
        raise NotImplementedError("mixed dislocations")


def prepare_shared_data(sample, settings, share):
    """Initialize and fill the shared containers.

    Parameters
    ----------
    sample : dict
        Dislocation sample data.
    settings : dict
        Diffraction variables.
    share : dict
        Shared data.

    """
    wrap = {}

    logger.debug("declaring shared data")
    share['x'] = RawArray('d', settings['n_all'])
    share['y'] = RawArray('d', settings['n_all'])
    share['s'] = RawArray('b', settings['n_all'])
    share['step'] = settings['s_vec_rf1'][0:2]
    share['steps'] = settings['steps']
    share['poisson'] = settings['poisson']
    share['constant'] = RawArray('d', 3)

    logger.debug("wrapping shared data")
    wrap['x'] = np.frombuffer(share['x'], dtype=np.float64)
    wrap['y'] = np.frombuffer(share['y'], dtype=np.float64)
    wrap['s'] = np.frombuffer(share['s'], dtype=np.int8)
    wrap['constant'] = np.frombuffer(share['constant'], dtype=np.float64)

    logger.debug("defining shared data")
    wrap['x'][0:settings['n_roi']] = sample['dislocations']['positions'][0]
    wrap['y'][0:settings['n_roi']] = sample['dislocations']['positions'][1]
    wrap['s'][0:settings['n_roi']] = sample['dislocations']['senses']
    wrap['constant'][:] = settings['g_vec_rf1'] * settings['b_len']
    wrap['constant'][1] *= - 1/(4*(1-settings['poisson']))
    if settings['replicate'] > 0:
        if sample['region']['type'] != 'square':
            raise TypeError("can only apply replication on a square region")
        for i, (k_x, k_y) in enumerate(shift_indexes(settings['replicate'])):
            j = (i+1) * settings['n_roi']
            k = (i+2) * settings['n_roi']
            wrap['x'][j:k] = k_x * sample['region']['side']
            wrap['y'][j:k] = k_y * sample['region']['side']
            wrap['x'][j:k] += wrap['x'][0:settings['n_roi']]
            wrap['y'][j:k] += wrap['y'][0:settings['n_roi']]
            wrap['s'][j:k] = wrap['s'][0:settings['n_roi']]


def user_check(settings, share, kwargs):
    """Display random points and replications if requested by the user.

    Parameters
    ----------
    settings : dict
        Diffraction variables.
    share : dict
        Shared data.
    kwargs : dict
        Keyword arguments passed to the main function.

    """
    logger.debug("checking random points and replicated dislocations")
    if kwargs['check']:
        fig, axes = plt.subplots()
        axes.scatter(share['x'], share['y'], label='dislocations')
        axes.scatter(*settings['points'].T, label='random points')
        axes.set_aspect(1)
        axes.set_xlabel("$x$ (m)")
        axes.set_ylabel("$y$ (m)")
        axes.legend()
        plt.show()
        plt.close(fig)


def assemble_data(sample, settings, kwargs, result, times):
    """Return the Fourier transform data.

    Parameters
    ----------
    sample : dict
        Dislocation sample data.
    settings : dict
        Diffraction variables.
    kwargs : dict
        Keyword arguments passed to the main function.
    results : dict
        Diffraction results.
    times : dict
        Time measurements.

    Returns
    -------
    dict
        Fourier transform data.

    """
    logger.debug("assembling data")
    data = {
        'distribution': sample['distribution'],
        'region': sample['region'],
        'diffraction': {
            'module': __package__,
            'z_uvw': kwargs['z_uvw'],
            'x_uvw': kwargs['x_uvw'],
            'b_uvw': kwargs['b_uvw'],
            'g_hkl': kwargs['g_hkl'],
            'type': settings['type'],
            'cell': kwargs['cell'],
            'step': kwargs['step'],
            'poisson': kwargs['poisson'],
            'contrast': settings['contrast'],
            'samples': 1,
            'replicate': kwargs['replicate'],
            'duration': times['1'] - times['0'],
            'points': kwargs['points'],
            'processes': kwargs['processes'],
        },
        'coefficients': {
            'harmonic': kwargs['harmonics'],
            'variable': result['variables'].tolist(),
            'cos_mean': result['avg_cos'].tolist(),
            'sin_mean': result['avg_sin'].tolist(),
            'cos_deviation': result['std_cos'].tolist(),
            'sin_deviation': result['std_sin'].tolist(),
        },
    }
    return data


def run_monte_carlo(settings, share, result, times):
    """Run the simulation with Monte Carlo method.

    Parameters
    ----------
    settings : dict
        Diffraction variables.
    share : dict
        Shared data.
    results : dict
        Diffraction results.
    times : dict
        Time measurements.

    """
    logger.debug("computing displacements")
    times['0'] = time()
    with Pool(
        settings['processes'],
        initializer=initializer,
        initargs=(share,),
    ) as pool:
        result['phases'] = pool.map(settings['function'], settings['points'])
    times['1'] = time()

    logger.debug("processing output")
    result['phases'] = np.einsum(
        'ij,k->ikj',
        result['phases'],
        settings['harmonics'],
    )
    result['cos'] = np.cos(result['phases'])
    result['sin'] = np.sin(result['phases'])
    result['avg_cos'] = np.average(result['cos'], axis=0)
    result['avg_sin'] = np.average(result['sin'], axis=0)
    result['coef'] = 1/np.sqrt(result['phases'].shape[0])
    result['std_cos'] = np.std(result['cos'], axis=0) * result['coef']
    result['std_sin'] = np.std(result['sin'], axis=0) * result['coef']
    result['variables'] = (np.arange(settings['steps'])+1) * settings['step']


def initializer(share):
    """Initialize a worker.

    Parameters
    ----------
    share : dict
        Shared data to be passed to the worker.

    """
    shared.update(share)


def shift_indexes(replications):
    """Return the indexes for the replications of the roi.

    Parameters
    ----------
    replications : int
        Number of replications around the region of interest.

    Returns
    -------
    list of tuple
        Replications indexes.

    """
    indexes = []
    for i in range(1, replications+1):
        for j in range(2*i):
            for k in (1, -1):
                indexes.append((-i*k, (i-j)*k))
                indexes.append(((i-j)*k,  i*k))
    return indexes


def random_points(sample, number):
    """Return the position of the random points.

    Parameters
    ----------
    sample : dict
        Dislocation sample data.
    number : int
        Number of random points in the region of interest.

    Returns
    -------
    np.array
        Position of the random points.

    """
    rng = np.random.default_rng(sample['distribution']['seed'])
    if sample['region']['type'] == 'square':
        side = sample['region']['side']
        positions = rng.random((number, 2), np.float64) * side
    elif sample['region']['type'] == 'disk':
        phi = 2 * np.pi * rng.random(number)
        rad = sample['region']['radius'] * np.sqrt(rng.random(number))
        return np.stack((rad*np.cos(phi), rad*np.sin(phi)), axis=1)
    else:
        raise NotImplementedError("unknown region type")
    return positions


def contrast_factor_edge(settings):
    """Return the contrast factor for edge dislocations.

    Parameters
    ----------
    settings : dict
        Diffraction settings.

    Returns
    -------
    float
        Contrast factor.

    """
    angle_z_g = settings['abs_ang_z_g']
    g_uni = settings['g_uni_rf2']
    b_uni = settings['b_uni_rf2']
    z_uni = settings['z_uni_rf2']
    poisson = settings['poisson']
    cos_projections = np.dot(
        unit_vector(g_uni - np.dot(g_uni, z_uni)*z_uni),
        unit_vector(b_uni - np.dot(b_uni, z_uni)*z_uni),
    )
    contrast = (
        np.sin(angle_z_g)**4 / (8*(1-poisson)**2)
        * (1-4*poisson+8*poisson**2+4*(1-2*poisson)*cos_projections**2)
    )
    return contrast


def contrast_factor_screw(settings):
    """Return the contrast factor for screw dislocations.

    Parameters
    ----------
    settings : dict
        Diffraction settings.

    Returns
    -------
    float
        Contrast factor.

    """
    angle = settings['abs_ang_z_g']
    return np.sin(angle)**2 * np.cos(angle)**2


def phases_screw(position):
    """Return the first harmonic phases at the position.

    This function applies only to screw dislocations. It returns
    'phase(L)' for each value 'L' of the Fourier variable. The
    Fourier coefficients for the harmonic 'h', and the variable 'L'
    can then be expressed by 'cos(h*phase(L))' and 'sin(h*phase(L))'.

    Parameters
    ----------
    position : np.array
        Position of the random point.

    Returns
    -------
    np.array
        Raw values.

    """
    step = shared['step']
    nsteps = shared['steps']
    constant = shared['constant']
    x_dislocations = shared['x']
    y_dislocations = shared['y']
    senses = shared['s']
    displacements = np.empty(nsteps+1, dtype=np.float64)
    for i in range(nsteps+1):
        x_measure, y_measure = position + i*step
        displacements[i] = displacement.screw(
            x_measure,
            y_measure,
            x_dislocations,
            y_dislocations,
            senses,
        )
    differences = displacements[1:] - displacements[0]
    phases = constant[2] * differences
    return phases


def phases_edge(position):
    """Return the first harmonic phases at the position.

    This function applies only to edge dislocations. It returns
    'phase(L)' for each value 'L' of the Fourier variable. The
    Fourier coefficients for the harmonic 'h', and the variable 'L'
    can then be expressed by 'cos(h*phase(L))' and 'sin(h*phase(L))'.

    Parameters
    ----------
    position : np.array
        Position of the random point.

    Returns
    -------
    np.array
        Raw values.

    """
    step = shared['step']
    nsteps = shared['steps']
    constant = shared['constant']
    x_dislocations = shared['x']
    y_dislocations = shared['y']
    senses = shared['s']
    poisson = shared['poisson']
    displacements = np.empty((nsteps+1, 2), dtype=np.float64)
    for i in range(nsteps+1):
        x_measure, y_measure = position + i*step
        displacements[i] = displacement.edge(
            x_measure,
            y_measure,
            x_dislocations,
            y_dislocations,
            senses,
            poisson,
        )
    differences = displacements[1:] - displacements[0]
    phases = np.inner(differences, constant[:2])
    return phases



# -*- coding: utf-8 -*-

"""Random dislocation drawing feature.

This module implements the functionality to randomly draw and save a
group of dislocation samples from the same probability distribution.

"""

from datetime import datetime, timezone
from logging import getLogger

from numlpa import config, kits
from numlpa._version import version_tuple
from numlpa.core.parallelism import get_rank, get_size
from numlpa.core.storage import Container, serialize


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.set_defaults(
        func=main,
    )
    parser.add_argument(
        'output_path',
        type=str,
        help="path of the output container",
    )
    parser.add_argument(
        '-n',
        '--size',
        type=int,
        help="number of samples to draw",
        metavar='int',
    )
    parser.add_argument(
        '-s',
        '--seed',
        type=int,
        help="random seed",
        metavar='int',
    )
    parser.add_argument(
        '-f',
        '--format',
        type=str,
        choices=('json', 'pyc'),
        help="format of the output files",
    )

    logger.debug("defining command-line subparsers")
    subparsers = parser.add_subparsers(
        dest='distribution',
        help="name of the module used for the random draw",
    )
    for name in kits.names('distributions'):
        subparser = subparsers.add_parser(
            name=name,
            help=f"draw from {name} distribution",
        )
        module = kits.get('distributions', name)
        module.setup(subparser)


def main(output_path, **kwargs):
    """Generate and save dislocation samples.

    Parameters
    ----------
    output_path : str
        Path of the output container.

    Keyword Arguments
    -----------------
    distribution : str
        Name of the module used for the random draw.
    size : int
        Number of samples to draw.
    seed : int
        Random seed.
    format : str
        Format of the output files.

    """
    containers = {}
    files = {}
    modules = {}

    logger.debug("retrieving parameters")
    kwargs.setdefault('distribution', config.get(__name__, 'distribution'))
    kwargs.setdefault('size', config.getint(__name__, 'size'))
    kwargs.setdefault('seed', config.getint(__name__, 'seed'))
    kwargs.setdefault('format', config.get(__name__, 'format'))

    logger.debug("preparing containers")
    containers['output'] = Container(output_path, create=True)

    logger.debug("preparing file names")
    stem_width = len(str(kwargs['size']-1))
    files['stems'] = [str(i).zfill(stem_width) for i in range(kwargs['size'])]
    files['output'] = [f"{stem}.{kwargs['format']}" for stem in files['stems']]
    files['existing'] = containers['output'].names()

    logger.debug("preparing submodules")
    modules['dist'] = kits.get('distributions', kwargs['distribution'])

    logger.debug("generating files")
    static = {
        'containers': containers,
        'files': files,
        'kwargs': kwargs,
        'modules': modules,
    }
    for i in range(get_rank(), kwargs['size'], get_size()):
        process(static, i)


def process(static, i):
    """Process a file.

    Parameters
    ----------
    static: dict of dict
        The data that does not vary from one file to another.
    i : int
        Index of the file to be generated.

    """
    data = {}
    file = {}
    passed = {}
    series = {}

    logger.debug("defining local variables for file index %d", i)
    file['output'] = static['files']['output'][i]
    file['target'] = f"({static['containers']['output']})[{file['output']}]"
    passed.update(static['kwargs'])
    passed['seed'] += i

    logger.debug("checking if %s already exists", file['output'])
    if file['output'] in static['files']['existing']:
        logger.debug("skipping %s", file['output'])
        return

    logger.info("generating %s", file['target'])
    data['output'] = {
        'metadata': {
            'type': 'dislocation-sample',
            'date': datetime.now(timezone.utc).isoformat(timespec="seconds"),
            'version': version_tuple,
        },
        **static['modules']['dist'].draw(**passed),
    }

    logger.debug("saving %s", file['output'])
    series['output'] = serialize(data['output'], file['output'])
    static['containers']['output'].add(series['output'], file['output'])



# -*- coding: utf-8 -*-

"""Parallelization on multiple machines.

This module implements the functionalities for retrieving the rank of
the machine and the number of machines allocated to the current task.

"""

from logging import getLogger, Formatter
from os import getenv

from numlpa import config


logger = getLogger(__name__)


def get_rank():
    """Return the rank of the machine.

    Returns
    -------
    int
        Rank of the machine.

    """
    logger.debug("retrieving rank")
    variable = config.get(__name__, 'rank')
    return int(getenv(variable, '0'))


def get_size():
    """Return the number of allocated machines.

    Returns
    -------
    int
        Number of allocated machines.

    """
    logger.debug("retrieving size")
    variable = config.get(__name__, 'size')
    return int(getenv(variable, '1'))


def get_format():
    """Return the format to use for logging on this machine.

    Returns
    -------
    str
        Logging format.

    """
    formatter = []
    formatter.append('%(levelname)s')
    size = get_size()
    if size > 1:
        width = len(str(size-1))
        formatter.append(str(get_rank()).zfill(width))
    formatter.append('%(name)s')
    formatter.append('%(message)s')
    return ':'.join(formatter)


getLogger().handlers[0].setFormatter(Formatter(get_format()))



from setuptools import Extension, setup

setup(
    ext_modules=[
        Extension(
            name="numlpa.kits.diffractometers.cpu.displacement",
            sources=["src/numlpa/kits/diffractometers/cpu/displacementmodule.c"],
        ),
        Extension(
            name="numlpa.kits.analyzers.cpu.interaction",
            sources=["src/numlpa/kits/analyzers/cpu/interactionmodule.c"],
        ),
    ]
)



# -*- coding: utf-8 -*-

"""Basic data representations.

This module implements very simple figure generators for the different
types of data produced by the package.

"""

from json import loads
from logging import getLogger

from matplotlib import patches
import matplotlib.pyplot as plt
import numpy as np

from numlpa import config


logger = getLogger(__package__)

TEX_G = r"\vec{g}"
TEX_U = r"\vec{u}"
TEX_R = r"\vec{r}"
TEX_L = r"\vec{L}"

CONVERSION = ['']


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.add_argument(
        '--figsize',
        help="width and heigth of the figure (inches)",
        type=float,
        nargs=2,
        metavar='float',
    )
    parser.add_argument(
        '--title',
        help="title of the figure",
        type=str,
        metavar='str',
    )
    legend = parser.add_mutually_exclusive_group()
    legend.add_argument(
        '--legend',
        action='store_const',
        const=True,
        help="display legends",
        dest='legend'
    )
    legend.add_argument(
        '--no-legend',
        action='store_const',
        const=False,
        help="hide legends",
        dest='legend'
    )
    parser.add_argument(
        '--length',
        help="length unit",
        type=str,
        choices=('m', 'nm'),
    )
    samples = parser.add_argument_group(
        title='sample',
        description="dislocation map options",
    )
    samples.add_argument(
        '--pairs',
        action='store_true',
        help="display a line between two consecutive dislocations",
    )
    samples.add_argument(
        '--subareas',
        help="name of the parameter giving the number of subareas",
        type=str,
    )
    profiles = parser.add_argument_group(
        title='profiles',
        description="diffraction profile options",
    )
    profiles.add_argument(
        '--ylog',
        action='store_true',
        help="display the ordinate axis in logarithmic scale",
    )


def retrieve_parameters(kwargs):
    """Load the default value of the missing parameters.

    Parameters
    ----------
    kwargs : dict
        Parameters passed to a function of the module.

    """
    logger.debug("retrieving parameters")
    kwargs.setdefault('figsize', loads(config.get(__name__, 'figsize')))
    kwargs.setdefault('title', config.get(__name__, 'title'))
    kwargs.setdefault('legend', config.getboolean(__name__, 'legend'))
    kwargs.setdefault('length', config.get(__name__, 'length'))
    kwargs.setdefault('pairs', config.getboolean(__name__, 'pairs'))
    kwargs.setdefault('subareas', config.get(__name__, 'subareas'))
    kwargs.setdefault('ylog', config.getboolean(__name__, 'ylog'))


def sample(data, **kwargs):
    """Return a matplotlib figure illustrating a dislocation sample.

    Parameters
    ----------
    data : dict
        Dislocation sample data.

    Keyword Arguments
    -----------------
    figsize : tuple
        Width and heigth of the figure (inches).
    title : str
        Title of the figure.
    legend : bool
        Display legends.
    length : str
        Length unit ('m' or 'nm').
    pairs : bool
        Display a line between two consecutive dislocations.
    subareas : str
        Name of the parameter giving the number of subareas.

    Returns
    -------
    Figure
        Matplotlib figure.

    """
    figure = {}
    retrieve_parameters(kwargs)

    logger.debug("initializing figure")
    figure['fig'], figure['ax'] = plt.subplots(figsize=kwargs['figsize'])
    figure['ax'].set_aspect(1)

    logger.debug("retrieving dislocation positions")
    figure['x'] = np.array(data['dislocations']['positions'][0])
    figure['y'] = np.array(data['dislocations']['positions'][1])
    figure['s'] = np.array(data['dislocations']['senses'])

    logger.debug("setting scale coefficient for unit conversion")
    figure['scale'] = 1
    if kwargs['length'] == 'nm':
        figure['scale'] = 1e9

    logger.debug("plotting dislocation positions")
    partition = (
        (r"$ \bot $", "Burgers vector sense $+$", figure['s'] > 0),
        (r"$ \top $", "Burgers vector sense $-$", figure['s'] < 0),
    )
    for marker, label, mask in partition:
        figure['ax'].scatter(
            figure['scale']*figure['x'][mask],
            figure['scale']*figure['y'][mask],
            marker=marker,
            label=label,
            zorder=100,
        )

    logger.debug("drawing region border")
    if data['region']['type'] == 'square':
        figure['side'] = figure['scale'] * data['region']['side']
        figure['limits'] = [-figure['side']*0.1, figure['side']*1.1]
        figure['ax'].set_xlim(figure['limits'])
        figure['ax'].set_ylim(figure['limits'])
        figure['nsub'] = data['distribution'].get(kwargs['subareas'], 1)
        for i in range(figure['nsub']+1):
            plt.plot(
                (0, figure['side']),
                (i*figure['side']/figure['nsub'],)*2,
                color='black',
                alpha=0.3,
            )
            plt.plot(
                (i*figure['side']/figure['nsub'],)*2,
                (0, figure['side']),
                color='black',
                alpha=0.3,
            )
    elif data['region']['type'] == 'disk':
        figure['radius'] = figure['scale'] * data['region']['radius']
        figure['limits'] = [-figure['radius']*1.1, figure['radius']*1.1]
        figure['ax'].set_xlim(figure['limits'])
        figure['ax'].set_ylim(figure['limits'])
        figure['circle'] = patches.Circle(
            (0, 0),
            figure['radius'],
            edgecolor='k',
            alpha=0.5,
            facecolor='none',
        )
        figure['ax'].add_patch(figure['circle'])

    logger.debug("displaying pairs")
    if kwargs['pairs']:
        for i in range(len(figure['s'])//2):
            figure['ax'].plot(
                (
                    figure['scale'] * figure['x'][2*i],
                    figure['scale'] * figure['x'][2*i+1],
                ),
                (
                    figure['scale'] * figure['y'][2*i],
                    figure['scale'] * figure['y'][2*i+1],
                ),
                color='black',
                alpha=0.3,
            )

    logger.debug("displaying text")
    if kwargs['legend']:
        figure['ax'].legend()
    if kwargs['title'] != "":
        figure['ax'].set_title(kwargs['title'])
    figure['ax'].set_xlabel(f"$x$ ({kwargs['length']})")
    figure['ax'].set_ylabel(f"$y$ ({kwargs['length']})")

    return figure['fig']


def transform(data, **kwargs):
    """Return a matplotlib figure illustrating a Fourier transform.

    Parameters
    ----------
    data : dict
        Fourier transform data.

    Keyword Arguments
    -----------------
    figsize : tuple
        Width and heigth of the figure (inches).
    title : str
        Title of the figure.
    legend : bool
        Display legends.
    length : str
        Length unit ('m' or 'nm').

    Returns
    -------
    Figure
        Matplotlib figure.

    """
    figure = {}
    retrieve_parameters(kwargs)

    logger.debug("initializing figure")
    figure['fig'], figure['ax'] = plt.subplots(figsize=kwargs['figsize'])
    figure['ax'].grid()
    if kwargs['ylog']:
        figure['ax'].set_yscale('log')

    logger.debug("setting scale coefficient for unit conversion")
    figure['scale'] = 1
    if kwargs['length'] == 'nm':
        figure['scale'] = 1e9

    logger.debug("plotting profiles")
    figure['variable'] = np.array(data['coefficients']['variable'])
    figure['cos'] = {
        'avg': np.array(data['coefficients']['cos_mean']),
        'std': np.array(data['coefficients']['cos_deviation']),
    }
    figure['sin'] = {
        'avg': np.array(data['coefficients']['sin_mean']),
        'std': np.array(data['coefficients']['sin_deviation']),
    }
    for func in ('cos', 'sin'):
        for i, j in enumerate(data['coefficients']['harmonic']):
            figure['ax'].plot(
                figure['scale']*figure['variable'],
                figure[func]['avg'][i],
                label=(
                    fr"$\left\langle \{func} \left( 2 \pi {TEX_G}_{j} \cdot "
                    fr"\left[ {TEX_U} \left( {TEX_R} + {TEX_L} \right) "
                    fr"- {TEX_U} \left( {TEX_R} \right) \right] \right) "
                    fr"\right\rangle$"
                )
            )
            figure['ax'].fill_between(
                figure['scale']*figure['variable'],
                figure[func]['avg'][i]-figure[func]['std'][i],
                figure[func]['avg'][i]+figure[func]['std'][i],
                alpha=0.3,
            )

    logger.debug("displaying text")
    if kwargs['legend']:
        figure['ax'].legend()
    if kwargs['title'] != "":
        figure['ax'].set_title(kwargs['title'])
    figure['ax'].set_xlabel(f"$L$ ({kwargs['length']})")

    return figure['fig']


def adjustment(data, **kwargs):
    """Return a matplotlib figure illustrating a Fourier transform.

    Parameters
    ----------
    data : dict
        Model adjustment data.

    Keyword Arguments
    -----------------
    figsize : tuple
        Width and heigth of the figure (inches).
    title : str
        Title of the figure.
    legend : bool
        Display legends.
    length : str
        Length unit ('m' or 'nm').

    Returns
    -------
    Figure
        Matplotlib figure.

    """
    figure = {}
    retrieve_parameters(kwargs)

    logger.debug("initializing figure")
    figure['fig'], figure['ax'] = plt.subplots(figsize=kwargs['figsize'])
    figure['ax'].grid()
    if kwargs['ylog']:
        figure['ax'].set_yscale('log')

    logger.debug("setting scale coefficient for unit conversion")
    figure['scale'] = 1
    if kwargs['length'] == 'nm':
        figure['scale'] = 1e9

    logger.debug("plotting model and simulated profile")
    figure['ax'].plot(
        np.array(data['profile']['variable'])*figure['scale'],
        np.array(data['profile']['amplitude']),
        label='profile',
    )
    figure['ax'].plot(
        np.array(data['model']['variable'])*figure['scale'],
        np.array(data['model']['amplitude']),
        label='model',
    )

    logger.debug("displaying text")
    if kwargs['legend']:
        figure['ax'].legend()
    if kwargs['title'] != "":
        figure['ax'].set_title(kwargs['title'])
    figure['ax'].set_xlabel(f"$L$ ({kwargs['length']})")

    return figure['fig']



# -*- coding: utf-8 -*-

"""Text fonts for figures.

This package gathers the different fonts available and implements the
functionalities to list the names and get the path of the files.

"""

from importlib import resources


def names():
    """Return the name of the available fonts.

    Returns
    -------
    list of str
        Names of the available fonts.

    """
    return []


def path(name):
    """Return the path to the font file.

    Parameters
    ----------
    name : str
        Name of the font.

    Returns
    -------
    str
        Path to the font file.

    """
    with resources.path(__package__, name) as file:
        file = str(file)
    return file



# -*- coding: utf-8 -*-

"""Image screw dislocation probability distribution.

This distribution allows to generate uniformly distributed screw
dislocations in a circular region of interest. The image dislocations
allowing the mechanical equilibrium at the surface of the cylinder are
also added outside the region of interest.

Below is an example of a set of parameter values for the distribution
and an illustration of a resulting dislocation sample.

* ``number`` = ``200`` (The region of interest contains 200
  dislocations.)

* ``radius`` = ``1000e-9`` (The region of interest is a disk of 1000 nm
  radius.)

.. image:: ../figures/distributions/images/0.svg
   :width: 600

"""

from logging import getLogger
from math import pi, sqrt, cos, sin
from random import choice, seed as set_seed, uniform

from numlpa import config


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.add_argument(
        '--number',
        type=int,
        help="number of dislocations in the region of interest",
        metavar='int',
    )
    parser.add_argument(
        '--radius',
        type=float,
        help="radius of the region of interest (m)",
        metavar='float',
    )


def draw(seed, **kwargs):
    """Return a sample drawn from the images distribution.

    Parameters
    ----------
    seed : int
        Random seed.

    Keyword Arguments
    -----------------
    number : int
        Number of dislocations in the region of interest.
    radius : float
        Radius of the region of interest (m).

    Returns
    -------
    dict
        Dislocation sample data.

    """
    logger.debug("retrieving parameters")
    kwargs.setdefault('number', config.getint(__name__, 'number'))
    kwargs.setdefault('radius', config.getfloat(__name__, 'radius'))
    density = kwargs['number']/(pi*kwargs['radius']**2)

    logger.debug("drawing dislocations")
    set_seed(seed)
    x_list, y_list, s_list = [], [], []
    for _ in range(kwargs['number']):
        angle = uniform(0, 2*pi)
        sense = choice((-1, 1))
        radius_dilocation = kwargs['radius'] * sqrt(uniform(0, 1))
        x_list.append(radius_dilocation*cos(angle))
        y_list.append(radius_dilocation*sin(angle))
        s_list.append(sense)
        radius_image = kwargs['radius']**2 / radius_dilocation
        x_list.append(radius_image*cos(angle))
        y_list.append(radius_image*sin(angle))
        s_list.append(-sense)

    logger.debug("assembling data")
    data = {
        'distribution': {
            'module': __name__,
            'seed': seed,
            'density': density,
            'number': kwargs['number'],
            'radius': kwargs['radius'],
        },
        'region': {
            'type': 'disk',
            'radius': kwargs['radius'],
        },
        'dislocations': {
            'senses': s_list,
            'positions': [x_list, y_list],
        },
    }
    return data



# -*- coding: utf-8 -*-

"""Groma-Ungár-Wilkens (GUW) model.

In 1988, I. Groma, T. Ungár and M. Wilkens [1]_ proposed a model that
has been simplified here as follows:

.. math::

   \\ln\\left[ A(L) \\right] = - \\Lambda \\rho L^2 \\left[
   \\ln\\left( \\frac{R_{eff}}{L} \\right) - \\frac{\\delta}{2}
   \\Lambda \\rho L^2 \\ln^2\\left( \\frac{R_0}{L} \\right) \\right]

With:

.. math::

   \\Lambda = \\frac{\\pi}{2} g^2 b^2 C_g

* :math:`A` Fourier amplitude.

* :math:`L` Fourier variable (:math:`\\mathrm{m}`).

* :math:`\\vec{g}` diffraction vector (:math:`\\mathrm{m}^{-1}`).

* :math:`\\vec{b}` Burgers vector (:math:`\\mathrm{m}`).

* :math:`C_g` contrast factor.

* :math:`\\rho` dislocation density (:math:`\\mathrm{m}^{-2}`).

* :math:`R_{eff}` effective radius (:math:`\\mathrm{m}`).

* :math:`\\delta` fluctuation.

* :math:`R_0` radius (:math:`\\mathrm{m}`).

For this model, the parameters to be adjusted are :math:`\\rho`,
:math:`R_{eff}`, :math:`\\delta` and :math:`R_0`.

.. [1] I. Groma, T. Ungár and M. Wilkens.
   Asymmetric X-ray line broadening of plastically deformed crystals.
   I. Theory, Journal of Applied Crystallography 21 (1) (1988) 47-54.

"""

from logging import getLogger

import numpy as np


logger = getLogger(__name__)


def guide(transform):
    """Return a guide for the model parameters.

    Parameters
    ----------
    transform : dict
        Fourier transform data.

    Returns
    -------
    dict
        Parameter names, min, max and real values.

    """
    logger.debug("retrieving distribution parameters")
    density = transform['distribution']['density']
    distance = 1/np.sqrt(density)

    logger.debug("assembling parameter guide")
    data = {
        'names': ('density', 'radius_eff', 'fluctuation', 'radius_0'),
        'max': (density*100, distance/10, 3, distance/10),
        'min': (density/100, distance*1000, 0, distance*1000),
        'real': (density, None, None, None),
    }
    return data


def model(transform, harmonic, limit=None):
    """Return the Fourier amplitude function.

    Parameters
    ----------
    transform : dict
        Fourier transform data.
    harmonic : int
        Harmonic of the diffraction vector.
    limit : int
        Index of the last Fourier variable.

    Returns
    -------
    Callable
        Function that takes parameter values and returns amplitudes.

    """
    data = {}

    logger.debug("defining constant quantities")
    data['L'] = np.array(transform['coefficients']['variable'])
    if limit:
        data['L'] = data['L'][:limit]
    data['b_uvw'] = np.array(transform['diffraction']['b_uvw'])
    data['g_hkl'] = np.array(transform['diffraction']['g_hkl'])
    data['g_vec'] = harmonic * data['g_hkl']/transform['diffraction']['cell']
    data['b_vec'] = data['b_uvw'] * transform['diffraction']['cell']/2
    data['g_len'] = np.linalg.norm(data['g_vec'])
    data['b_len'] = np.linalg.norm(data['b_vec'])
    data['C'] = transform['diffraction']['contrast']
    data['Lambda'] = np.pi/2 * data['g_len']**2 * data['b_len']**2 * data['C']
    data['cst1'] = data['Lambda'] * data['L']**2

    def amplitude(density, radius_eff, fluctuation, radius_0):
        if density < 0 or radius_eff < 0 or radius_0 < 0:
            return np.inf
        term_1 = data['cst1'] * density
        term_2 = np.log(radius_eff / data['L'])
        term_3 = np.log(radius_0 / data['L'])
        return np.exp(-term_1*(term_2 - fluctuation/2*term_1*term_3**2))

    return amplitude
